from __future__ import annotations

__all__ = (
    "InlineQueryResult",
    "InlineQueryResultArticle",
    "InlineQueryResultAudio",
    "InlineQueryResultsButton",
    "InlineQueryResultCachedAudio",
    "InlineQueryResultCachedDocument",
    "InlineQueryResultCachedGif",
    "InlineQueryResultCachedMpeg4Gif",
    "InlineQueryResultCachedPhoto",
    "InlineQueryResultCachedSticker",
    "InlineQueryResultCachedVideo",
    "InlineQueryResultCachedVoice",
    "InlineQueryResultContact",
    "InlineQueryResultDocument",
    "InlineQueryResultGame",
    "InlineQueryResultGif",
    "InlineQueryResultLocation",
    "InlineQueryResultMpeg4Gif",
    "InlineQueryResultPhoto",
    "InlineQueryResultVenue",
    "InlineQueryResultVideo",
    "InlineQueryResultVoice",
)

from typing import Literal
from typing import TypeAlias
from typing import final

from msgspec import field

from yatbaf.enums import DocumentMimeType
from yatbaf.enums import ParseMode
from yatbaf.enums import ThumbnailMimeType
from yatbaf.enums import VideoMimeType
from yatbaf.typing import NoneBool
from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from .abc import TelegramType
from .inline_keyboard_markup import InlineKeyboardMarkup
from .input_message_content import InputMessageContent
from .message_entity import MessageEntity
from .web_app_info import WebAppInfo


@final
class InlineQueryResultGame(TelegramType):
    """Represents a Game.

    See: https://core.telegram.org/bots/api#inlinequeryresultgame
    """

    id: str
    """Unique identifier for this result, 1-64 bytes."""

    game_short_name: str
    """Short name of the game."""

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    type: Literal["game"] = field(default_factory=lambda: "game")
    """Type of the result, must be `game`."""


@final
class InlineQueryResultPhoto(TelegramType):
    """Represents a link to a photo stored on the Telegram servers. By default,
    this photo will be sent by the user with an optional caption. Alternatively,
    you can use ``input_message_content`` to send a message with the specified
    content instead of the photo.

    See: https://core.telegram.org/bots/api#inlinequeryresultphoto
    """

    id: str
    """Unique identifier for this result, 1-64 bytes."""

    photo_url: str
    """A valid URL of the photo. Photo must be in JPEG format. Photo size must
    not exceed 5MB.
    """

    thumbnail_url: str
    """URL of the thumbnail for the photo."""

    photo_width: NoneInt = None
    """*Optional.* Width of the photo."""

    photo_height: NoneInt = None
    """*Optional.* Height of the photo."""

    title: NoneStr = None
    """*Optional.* Title for the result."""

    description: NoneStr = None
    """*Optional.* Short description of the result."""

    caption: NoneStr = None
    """*Optional.* Caption of the photo to be sent, 0-1024 characters after
    entities parsing.
    """

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the photo caption. See
    formatting options for more details.
    """

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of ``parse_mode``.
    """

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    input_message_content: InputMessageContent | None = None
    """*Optional.* Content of the message to be sent instead of the photo."""

    type: Literal["photo"] = field(default_factory=lambda: "photo")
    """Type of the result, must be `photo`."""


@final
class InlineQueryResultAudio(TelegramType):
    """Represents a link to an MP3 audio file. By default, this audio file will
    be sent by the user. Alternatively, you can use ``input_message_content``
    to send a message with the specified content instead of the audio.

    See: https://core.telegram.org/bots/api#inlinequeryresultaudio
    """

    id: str
    """Unique identifier for this result, 1-64 bytes."""

    audio_url: str
    """A valid URL for the audio file."""

    title: str
    """Title."""

    caption: str | None = None
    """*Optional.* Caption, 0-1024 characters after entities parsing."""

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the audio caption."""

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of ``parse_mode``.
    """

    performer: NoneStr = None
    """*Optional.* Performer."""

    audio_duration: NoneInt = None
    """*Optional.* Audio duration in seconds."""

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    input_message_content: InputMessageContent | None = None
    """*Optional.* Content of the message to be sent instead of the audio."""

    type: Literal["audio"] = field(default_factory=lambda: "audio")
    """Type of the result, must be `audio`."""


@final
class InlineQueryResultVoice(TelegramType):
    """Represents a link to a voice recording in an .OGG container encoded with
    OPUS. By default, this voice recording will be sent by the user.
    Alternatively, you can use ``input_message_content`` to send a message with
    the specified content instead of the the voice message.

    See: https://core.telegram.org/bots/api#inlinequeryresultvoice
    """

    id: str
    """Unique identifier for this result, 1-64 bytes."""

    voice_url: str
    """A valid URL for the voice recording."""

    title: str
    """Recording title."""

    caption: NoneStr = None
    """*Optional.* Caption, 0-1024 characters after entities parsing."""

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the voice message caption."""

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of ``parse_mode``."""

    voice_duration: NoneInt = None
    """*Optional.* Recording duration in seconds."""

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    input_message_content: InputMessageContent | None = None
    """*Optional.* Content of the message to be sent instead of the voice
    recording.
    """

    type: Literal["voice"] = field(default_factory=lambda: "voice")
    """Type of the result, must be `voice`."""


@final
class InlineQueryResultVenue(TelegramType):
    """Represents a venue. By default, the venue will be sent by the user.
    Alternatively, you can use ``input_message_content`` to send a message with
    the specified content instead of the venue.

    See: https://core.telegram.org/bots/api#inlinequeryresultvenue
    """

    id: str
    """Unique identifier for this result, 1-64 Bytes."""

    latitude: float
    """Latitude of the venue location in degrees."""

    longitude: float
    """Longitude of the venue location in degrees."""

    address: str
    """Address of the venue."""

    thumbnail_url: str
    """*Optional.* Url of the thumbnail for the result."""

    title: NoneStr = None
    """Title of the venue."""

    foursquare_id: NoneInt = None
    """*Optional.* Foursquare identifier of the venue if known."""

    foursquare_type: NoneStr = None
    """*Optional.* Foursquare type of the venue, if known."""

    google_place_id: NoneStr = None
    """*Optional.* Google Places identifier of the venue."""

    google_place_type: NoneStr = None
    """*Optional.* Google Places type of the venue."""

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    input_message_content: InputMessageContent | None = None
    """*Optional.* Content of the message to be sent instead of the venue."""

    thumbnail_width: NoneInt = None
    """*Optional.* Thumbnail width."""

    thumbnail_height: NoneInt = None
    """*Optional.* Thumbnail height."""

    type: Literal["venue"] = field(default_factory=lambda: "venue")
    """Type of the result, must be `venue`."""


@final
class InlineQueryResultVideo(TelegramType):
    """Represents a link to a page containing an embedded video player or a
    video file. By default, this video file will be sent by the user with an
    optional caption. Alternatively, you can use ``input_message_content`` to
    send a message with the specified content instead of the video.

    .. note::

        If an InlineQueryResultVideo message contains an embedded video
        (e.g., YouTube), you *must* replace its content using
        ``input_message_content``.

    See: https://core.telegram.org/bots/api#inlinequeryresultvideo
    """

    id: str
    """Unique identifier for this result, 1-64 bytes."""

    video_url: str
    """A valid URL for the embedded video player or video file."""

    mime_type: VideoMimeType
    """MIME type of the content of the video URL, 'text/html' or 'video/mp4'."""

    thumbnail_url: str
    """URL of the thumbnail (JPEG only) for the video."""

    title: NoneStr = None
    """Title for the result."""

    caption: NoneStr = None
    """*Optional.* Caption of the video to be sent, 0-1024 characters after
    entities parsing.
    """

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the video caption. See
    formatting options for more details.
    """

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of ``parse_mode``.
    """

    video_width: NoneInt = None
    """*Optional.* Video width."""

    video_height: NoneInt = None
    """*Optional.* Video height."""

    video_duration: NoneInt = None
    """*Optional.* Video duration in seconds."""

    description: NoneStr = None
    """*Optional.* Short description of the result."""

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    input_message_content: InputMessageContent | None = None
    """*Optional.* Content of the message to be sent instead of the video. This
    field is *required* if InlineQueryResultVideo is used to send an HTML-page
    as a result (e.g., a YouTube video).
    """

    type: Literal["video"] = field(default_factory=lambda: "video")
    """Type of the result, must be `video`."""


@final
class InlineQueryResultArticle(TelegramType):
    """Represents a link to an article or web page.

    See: https://core.telegram.org/bots/api#inlinequeryresultarticle
    """

    id: str
    """Unique identifier for this result, 1-64 Bytes."""

    title: str
    """Title of the result."""

    input_message_content: InputMessageContent
    """Content of the message to be sent."""

    reply_markup: InlineKeyboardMarkup | None = None
    """Optional. Inline keyboard attached to the message."""

    url: NoneStr = None
    """*Optional.* URL of the result."""

    hide_url: NoneBool = None
    """*Optional.* Pass True if you don't want the URL to be shown in the
    message.
    """

    description: NoneStr = None
    """*Optional.* Short description of the result."""

    thumbnail_url: NoneStr = None
    """*Optional.* Url of the thumbnail for the result."""

    thumbnail_width: NoneInt = None
    """*Optional.* Thumbnail width."""

    thumbnail_height: NoneInt = None
    """*Optional.* Thumbnail height."""

    type: Literal["article"] = field(default_factory=lambda: "article")
    """Type of the result, must be `article`."""


@final
class InlineQueryResultContact(TelegramType):
    """Represents a contact with a phone number. By default, this contact will
    be sent by the user. Alternatively, you can use ``input_message_content``
    to send a message with the specified content instead of the contact.

    See: https://core.telegram.org/bots/api#inlinequeryresultcontact
    """

    id: str
    """Unique identifier for this result, 1-64 Bytes."""

    phone_number: str
    """Contact's phone number."""

    first_name: str
    """Contact's first name."""

    thumbnail_url: str
    """*Optional.* Url of the thumbnail for the result."""

    last_name: NoneStr = None
    """*Optional.* Contact's last name."""

    vcard: NoneStr = None
    """*Optional.* Additional data about the contact in the form of a vCard,
    0-2048 bytes.
    """

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    input_message_content: InputMessageContent | None = None
    """*Optional.* Content of the message to be sent instead of the contact."""

    thumbnail_width: NoneInt = None
    """*Optional.* Thumbnail width."""

    thumbnail_height: NoneInt = None
    """*Optional.* Thumbnail height."""

    type: Literal["contact"] = field(default_factory=lambda: "contact")
    """Type of the result, must be `contact`."""


@final
class InlineQueryResultLocation(TelegramType):
    """Represents a location on a map. By default, the location will be sent
    by the user. Alternatively, you can use ``input_message_content`` to send
    a message with the specified content instead of the location.

    See: https://core.telegram.org/bots/api#inlinequeryresultlocation
    """

    id: str
    """Unique identifier for this result, 1-64 Bytes."""

    latitude: float
    """Location latitude in degrees."""

    longitude: float
    """Location longitude in degrees."""

    thumbnail_url: str
    """*Optional.* Url of the thumbnail for the result."""

    title: NoneStr = None
    """Location title."""

    horizontal_accuracy: float | None = None
    """*Optional.* The radius of uncertainty for the location, measured in
    meters; 0-1500.
    """

    live_period: int | None = None
    """*Optional.* Period in seconds for which the location can be updated,
    should be between 60 and 86400.
    """

    heading: int | None = None
    """*Optional.* For live locations, a direction in which the user is moving,
    in degrees. Must be between 1 and 360 if specified.
    """

    proximity_alert_radius: int | None = None
    """*Optional.* For live locations, a maximum distance for proximity alerts
    about approaching another chat member, in meters. Must be between 1 and
    100000 if specified.
    """

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    input_message_content: InputMessageContent | None = None
    """*Optional.* Content of the message to be sent instead of the location."""

    thumbnail_width: NoneInt = None
    """*Optional.* Thumbnail width."""

    thumbnail_height: NoneInt = None
    """*Optional.* Thumbnail height."""

    type: Literal["location"] = field(default_factory=lambda: "location")
    """Type of the result, must be `location`."""


@final
class InlineQueryResultDocument(TelegramType):
    """Represents a link to a file. By default, this file will be sent by the
    user with an optional caption. Alternatively, you can use
    ``input_message_content`` to send a message with the specified content
    instead of the file. Currently, only .PDF and .ZIP files can be sent using
    this method.

    See: https://core.telegram.org/bots/api#inlinequeryresultdocument
    """

    id: str
    """Unique identifier for this result, 1-64 bytes."""

    title: str
    """Title for the result."""

    document_url: str
    """A valid URL for the file."""

    mime_type: DocumentMimeType
    """MIME type of the content of the file, either 'application/pdf' or
    'application/zip'.
    """

    caption: str | None = None
    """*Optional.* Caption of the document to be sent, 0-1024 characters after
    entities parsing.
    """

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the document caption."""

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of ``parse_mode``.
    """

    description: NoneStr = None
    """*Optional.* Short description of the result."""

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    input_message_content: InputMessageContent | None = None
    """*Optional.* Content of the message to be sent instead of the file."""

    thumbnail_url: NoneStr = None
    """*Optional.* URL of the thumbnail (JPEG only) for the file."""

    thumbnail_width: NoneInt = None
    """*Optional.* Thumbnail width."""

    thumbnail_height: NoneInt = None
    """*Optional.* Thumbnail height."""

    type: Literal["document"] = field(default_factory=lambda: "document")
    """Type of the result, must be `document`."""


@final
class InlineQueryResultMpeg4Gif(TelegramType):
    """Represents a link to a video animation (H.264/MPEG-4 AVC video without
    sound). By default, this animated MPEG-4 file will be sent by the user with
    optional caption. Alternatively, you can use ``input_message_content``
    to send a message with the specified content instead of the animation.

    See: https://core.telegram.org/bots/api#inlinequeryresultmpeg4gif
    """

    id: str
    """Unique identifier for this result, 1-64 bytes."""

    mpeg4_url: str
    """A valid URL for the MPEG4 file. File size must not exceed 1MB."""

    thumbnail_url: str
    """URL of the static (JPEG or GIF) or animated (MPEG4) thumbnail for the
    result.
    """

    mpeg4_width: NoneInt = None
    """Optional. Video width."""

    mpeg4_height: NoneInt = None
    """Optional. Video height."""

    mpeg4_duration: NoneInt = None
    """Optional. Video duration in seconds."""

    thumbnail_mime_type: ThumbnailMimeType | None = None
    """*Optional.* MIME type of the thumbnail, must be one of 'image/jpeg',
    'image/gif', or 'video/mp4'. Defaults to 'image/jpeg'.
    """

    title: NoneStr = None
    """Optional. Title for the result."""

    caption: NoneStr = None
    """*Optional.* Caption of the MPEG-4 file to be sent, 0-1024 characters
    after entities parsing.
    """

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the caption."""

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of ``parse_mode``.
    """

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    input_message_content: InputMessageContent | None = None
    """*Optional.* Content of the message to be sent instead of the video
    animation.
    """

    type: Literal["mpeg4_gif"] = field(default_factory=lambda: "mpeg4_gif")
    """Type of the result, must be `mpeg4_gif`."""


@final
class InlineQueryResultCachedGif(TelegramType):
    """Represents a link to an animated GIF file stored on the Telegram servers.
    By default, this animated GIF file will be sent by the user with an optional
    caption. Alternatively, you can use ``input_message_content`` to send a
    message with specified content instead of the animation.

    See: https://core.telegram.org/bots/api#inlinequeryresultcachedgif
    """

    id: str
    """Unique identifier for this result, 1-64 bytes."""

    gif_file_id: str
    """A valid file identifier for the GIF file."""

    title: NoneStr = None
    """*Optional.* Title for the result."""

    caption: str | None = None
    """*Optional.* Caption of the GIF file to be sent, 0-1024 characters after
    entities parsing.
    """

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the caption."""

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of ``parse_mode``.
    """

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    input_message_content: InputMessageContent | None = None
    """*Optional.* Content of the message to be sent instead of the GIF
    animation.
    """

    type: Literal["gif"] = field(default_factory=lambda: "gif")
    """Type of the result, must be `gif`."""


@final
class InlineQueryResultCachedPhoto(TelegramType):
    """Represents a link to a photo stored on the Telegram servers. By default,
    this photo will be sent by the user with an optional caption. Alternatively,
    you can use ``input_message_content`` to send a message with the specified
    content instead of the photo.

    See: https://core.telegram.org/bots/api#inlinequeryresultcachedphoto
    """

    id: str
    """Unique identifier for this result, 1-64 bytes."""

    photo_file_id: str
    """A valid file identifier of the photo."""

    title: NoneStr = None
    """*Optional.* Title for the result."""

    description: NoneStr = None
    """*Optional.* Short description of the result."""

    caption: str | None = None
    """*Optional.* Caption of the photo to be sent, 0-1024 characters after
    entities parsing.
    """

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the photo caption."""

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of ``parse_mode``.
    """

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    input_message_content: InputMessageContent | None = None
    """*Optional.* Content of the message to be sent instead of the photo."""

    type: Literal["photo"] = field(default_factory=lambda: "photo")
    """Type of the result, must be photo."""


@final
class InlineQueryResultCachedVoice(TelegramType):
    """Represents a link to a voice message stored on the Telegram servers. By
    default, this voice message will be sent by the user. Alternatively, you can
    use ``input_message_content`` to send a message with the specified content
    instead of the voice message.

    See: https://core.telegram.org/bots/api#inlinequeryresultcachedvoice
    """

    id: str
    """Unique identifier for this result, 1-64 bytes."""

    voice_file_id: str
    """A valid file identifier for the voice message."""

    title: str
    """Voice message title."""

    caption: str | None = None
    """*Optional.* Caption, 0-1024 characters after entities parsing."""

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the voice message caption."""

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of ``parse_mode``.
    """

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    input_message_content: InputMessageContent | None = None
    """*Optional.* Content of the message to be sent instead of the voice
    message.
    """

    type: Literal["voice"] = field(default_factory=lambda: "voice")
    """Type of the result, must be `voice`."""


@final
class InlineQueryResultCachedAudio(TelegramType):
    """Represents a link to an MP3 audio file stored on the Telegram servers.
    By default, this audio file will be sent by the user. Alternatively, you can
    use ``input_message_content`` to send a message with the specified content
    instead of the audio.

    See: https://core.telegram.org/bots/api#inlinequeryresultcachedaudio
    """

    id: str
    """Unique identifier for this result, 1-64 bytes."""

    audio_file_id: str
    """A valid file identifier for the audio file."""

    caption: str | None = None
    """*Optional.* Caption, 0-1024 characters after entities parsing."""

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the audio caption."""

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of parse_mode.
    """

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    input_message_content: InputMessageContent | None = None
    """*Optional.* Content of the message to be sent instead of the audio."""

    type: Literal["audio"] = field(default_factory=lambda: "audio")
    """Type of the result, must be `audio`."""


@final
class InlineQueryResultCachedVideo(TelegramType):
    """Represents a link to a video file stored on the Telegram servers. By
    default, this video file will be sent by the user with an optional caption.
    Alternatively, you can use ``input_message_content`` to send a message with
    the specified content instead of the video.

    See: https://core.telegram.org/bots/api#inlinequeryresultcachedvideo
    """

    id: str
    """Unique identifier for this result, 1-64 bytes."""

    video_file_id: str
    """A valid file identifier for the video file."""

    title: NoneStr = None
    """Title for the result."""

    description: NoneStr = None
    """*Optional.* Short description of the result."""

    caption: str | None = None
    """*Optional.* Caption of the video to be sent, 0-1024 characters after
    entities parsing.
    """

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the video caption."""

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of parse_mode.
    """

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    input_message_content: InputMessageContent | None = None
    """*Optional.* Content of the message to be sent instead of the video."""

    type: Literal["video"] = field(default_factory=lambda: "video")
    """Type of the result, must be `video`."""


@final
class InlineQueryResultCachedSticker(TelegramType):
    """Represents a link to a sticker stored on the Telegram servers. By
    default, this sticker will be sent by the user. Alternatively, you can use
    ``input_message_content`` to send a message with the specified content
    instead of the sticker.

    See: https://core.telegram.org/bots/api#inlinequeryresultcachedsticker
    """

    id: str
    """Unique identifier for this result, 1-64 bytes."""

    sticker_file_id: str
    """A valid file identifier of the sticker."""

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    input_message_content: InputMessageContent | None = None
    """*Optional.* Content of the message to be sent instead of the sticker."""

    type: Literal["sticker"] = field(default_factory=lambda: "sticker")
    """Type of the result, must be `sticker`."""


@final
class InlineQueryResultCachedDocument(TelegramType):
    """Represents a link to a file stored on the Telegram servers. By default,
    this file will be sent by the user with an optional caption. Alternatively,
    you can use ``input_message_content`` to send a message with the specified
    content instead of the file.

    See: https://core.telegram.org/bots/api#inlinequeryresultcacheddocument
    """

    id: str
    """Unique identifier for this result, 1-64 bytes."""

    title: str
    """Title for the result."""

    document_file_id: str
    """A valid file identifier for the file."""

    description: NoneStr = None
    """*Optional.* Short description of the result."""

    caption: str | None = None
    """*Optional.* Caption of the document to be sent, 0-1024 characters after
    entities parsing.
    """

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the document caption."""

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of ``parse_mode``.
    """

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    input_message_content: InputMessageContent | None = None
    """*Optional.* Content of the message to be sent instead of the file."""

    type: Literal["document"] = field(default_factory=lambda: "document")
    """Type of the result, must be `document`."""


@final
class InlineQueryResultCachedMpeg4Gif(TelegramType):
    """Represents a link to a video animation (H.264/MPEG-4 AVC video without
    sound) stored on the Telegram servers. By default, this animated MPEG-4 file
    will be sent by the user with an optional caption. Alternatively, you can
    use ``input_message_content`` to send a message with the specified content
    instead of the animation.

    See: https://core.telegram.org/bots/api#inlinequeryresultcachedmpeg4gif
    """

    id: str
    """Unique identifier for this result, 1-64 bytes."""

    mpeg4_file_id: str
    """A valid file identifier for the MPEG4 file."""

    title: NoneStr = None
    """Optional. Title for the result."""

    caption: str | None = None
    """*Optional.* Caption of the MPEG-4 file to be sent, 0-1024 characters
    after entities parsing.
    """

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the caption."""

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of ``parse_mode``.
    """

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    input_message_content: InputMessageContent | None = None
    """*Optional.* Content of the message to be sent instead of the video
    animation.
    """

    type: Literal["mpeg4_gif"] = field(default_factory=lambda: "mpeg4_gif")
    """Type of the result, must be `mpeg4_gif`."""


@final
class InlineQueryResultGif(TelegramType):
    """Represents a link to an animated GIF file. By default, this animated GIF
    file will be sent by the user with optional caption. Alternatively, you can
    use ``input_message_content`` to send a message with the specified content
    instead of the animation.

    See: https://core.telegram.org/bots/api#inlinequeryresultgif
    """

    id: str
    """Unique identifier for this result, 1-64 bytes."""

    gif_url: str
    """A valid URL for the GIF file. File size must not exceed 1MB."""

    thumbnail_url: str
    """URL of the static (JPEG or GIF) or animated (MPEG4) thumbnail for the
    result.
    """

    gif_width: NoneInt = None
    """Optional. Width of the GIF."""

    gif_height: NoneInt = None
    """Optional. Height of the GIF."""

    gif_duration: NoneInt = None
    """Optional. Duration of the GIF in seconds."""

    thumbnail_mime_type: ThumbnailMimeType | None = None
    """*Optional.* MIME type of the thumbnail."""

    title: NoneStr = None
    """*Optional.* Title for the result."""

    caption: str | None = None
    """*Optional.* Caption of the GIF file to be sent, 0-1024 characters after
    entities parsing.
    """

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the caption."""

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of ``parse_mode``.
    """

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message."""

    input_message_content: InputMessageContent | None = None
    """*Optional.* Content of the message to be sent instead of the GIF
    animation.
    """

    type: Literal["gif"] = field(default_factory=lambda: "gif")
    """Type of the result, must be `gif`."""


@final
class InlineQueryResultsButton(TelegramType):
    """This object represents a button to be shown above inline query results.
    You *must* use exactly one of the optional fields.

    See: https://core.telegram.org/bots/api#inlinequeryresultsbutton
    """

    text: str
    """Label text on the button."""

    web_app: WebAppInfo | None = None
    """*Optional.* Description of the Web App that will be launched when the
    user presses the button. The Web App will be able to switch back to the
    inline mode using the method switchInlineQuery inside the Web App.
    """

    start_parameter: NoneStr = None
    """*Optional.* Deep-linking parameter for the /start message sent to the
    bot when a user presses the button. 1-64 characters, only A-Z, a-z, 0-9, _
    and - are allowed.
    """


# https://core.telegram.org/bots/api#inlinequeryresult
InlineQueryResult: TypeAlias = (
    "InlineQueryResultArticle "
    "| InlineQueryResultAudio "
    "| InlineQueryResultCachedAudio "
    "| InlineQueryResultCachedDocument "
    "| InlineQueryResultCachedGif "
    "| InlineQueryResultCachedMpeg4Gif "
    "| InlineQueryResultCachedPhoto "
    "| InlineQueryResultCachedSticker "
    "| InlineQueryResultCachedVideo "
    "| InlineQueryResultCachedVoice "
    "| InlineQueryResultContact "
    "| InlineQueryResultDocument "
    "| InlineQueryResultGame "
    "| InlineQueryResultGif "
    "| InlineQueryResultLocation "
    "| InlineQueryResultMpeg4Gif "
    "| InlineQueryResultPhoto "
    "| InlineQueryResultVenue "
    "| InlineQueryResultVideo "
    "| InlineQueryResultVoice"
)



from __future__ import annotations

__all__ = ("BackgroundTasks",)

import asyncio
import logging
from typing import TYPE_CHECKING
from typing import ParamSpec
from typing import final

if TYPE_CHECKING:
    from collections.abc import Coroutine

    from .typing import ErrorHandler

log = logging.getLogger(__name__)

P = ParamSpec("P")


@final
class BackgroundTasks:
    """Task manager."""

    __slots__ = (
        "_tasks",
        "_error_handler",
    )

    def __init__(self, error_handler: ErrorHandler, /) -> None:
        """
        :param error_handler: Error handler.
        """
        self._tasks: set[asyncio.Task] = set()
        self._error_handler = error_handler

    async def _wrapper(self, coro: Coroutine) -> None:
        try:
            await coro
        except Exception as error:
            await self._error_handler.on_error(error)

    def _create_task(self, coro: Coroutine, *, name: str | None = None) -> None:
        task = asyncio.create_task(coro, name=name)
        self._tasks.add(task)
        task.add_done_callback(self._tasks.discard)

    def add(self, coro: Coroutine, *, name: str | None = None) -> None:
        """Wrap the coroutine into a :class:`~asyncio.Task` and add to manager.

        :param coro: Coroutine to run as task.
        :param name: *Optional.* Task name.
        """
        self._create_task(self._wrapper(coro), name=name)

    @staticmethod
    async def kill(task: asyncio.Task, /) -> None:
        """Cancel task.

        :param task: :class:`~asyncio.Task` to be canceled.
        """
        if not task.done():
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass

    async def kill_all(self) -> None:
        """Cancel all tasks stored in manager."""
        while self._tasks:
            task = self._tasks.pop()
            await self.kill(task)

    async def wait_all(self) -> None:
        """Wait for all active tasks stored in manager."""
        if self._tasks:
            await asyncio.gather(*self._tasks)



import pytest

from yatbaf.enums import UpdateType
from yatbaf.middleware import Middleware
from yatbaf.router import OnCallbackQuery
from yatbaf.router import OnChannelPost
from yatbaf.router import OnChatBoost
from yatbaf.router import OnChatJoinRequest
from yatbaf.router import OnChatMemeber
from yatbaf.router import OnChosenInlineResult
from yatbaf.router import OnEditedChannelPost
from yatbaf.router import OnEditedMessage
from yatbaf.router import OnInlineQuery
from yatbaf.router import OnMessage
from yatbaf.router import OnMessageReaction
from yatbaf.router import OnMessageReactionCount
from yatbaf.router import OnMyChatMember
from yatbaf.router import OnPoll
from yatbaf.router import OnPollAnswer
from yatbaf.router import OnPreCheckoutQuery
from yatbaf.router import OnRemovedChatBoost
from yatbaf.router import OnShippingQuery
from yatbaf.router import router_map


def test_router_map():
    for type_ in UpdateType:
        assert type_ in router_map


def test_update_type():
    assert OnCallbackQuery._update_type == UpdateType.CALLBACK_QUERY
    assert OnChannelPost._update_type == UpdateType.CHANNEL_POST
    assert OnChatJoinRequest._update_type == UpdateType.CHAT_JOIN_REQUEST
    assert OnChatMemeber._update_type == UpdateType.CHAT_MEMBER
    assert OnChosenInlineResult._update_type == UpdateType.CHOSEN_INLINE_RESULT
    assert OnEditedChannelPost._update_type == UpdateType.EDITED_CHANNEL_POST
    assert OnEditedMessage._update_type == UpdateType.EDITED_MESSAGE
    assert OnInlineQuery._update_type == UpdateType.INLINE_QUERY
    assert OnMessage._update_type == UpdateType.MESSAGE
    assert OnMyChatMember._update_type == UpdateType.MY_CHAT_MEMBER
    assert OnPoll._update_type == UpdateType.POLL
    assert OnPollAnswer._update_type == UpdateType.POLL_ANSWER
    assert OnPreCheckoutQuery._update_type == UpdateType.PRE_CHECKOUT_QUERY
    assert OnShippingQuery._update_type == UpdateType.SHIPPING_QUERY
    assert OnChatBoost._update_type == UpdateType.CHAT_BOOST
    assert OnRemovedChatBoost._update_type == UpdateType.REMOVED_CHAT_BOOST
    assert OnMessageReaction._update_type == UpdateType.MESSAGE_REACTION
    assert OnMessageReactionCount._update_type == UpdateType.MESSAGE_REACTION_COUNT  # noqa: E501


def test_new_router():
    router = OnMessage()
    assert not router._handlers
    assert not router._middleware
    assert not router._guards


@pytest.mark.parametrize(
    "objs",
    [
        (
            OnMessage(),
            OnMessage(),
        ),
        (
            OnMessage(guards=[g := object()]),
            OnMessage(guards=[g]),
        ),
        (
            OnMessage(handlers=[h := object()]),
            OnMessage(handlers=[h]),
        ),
        (
            OnMessage(guards=[g := object()], handlers=[h := object()]),
            OnMessage(guards=[g], handlers=[h]),
        ),
        (
            OnMessage(
                middleware=[m := object()],
                guards=[g := object()],
                handlers=[h := object()]
            ),
            OnMessage(
                middleware=[m],
                guards=[g],
                handlers=[h],
            ),
        ),
        (
            OnMessage(
                middleware=[Middleware(m := object())],
                guards=[g := object()],
                handlers=[h := object()]
            ),
            OnMessage(
                middleware=[Middleware(m)],
                guards=[g],
                handlers=[h],
            ),
        ),
    ]
)
def test_eq(objs):
    router1, router2 = objs
    assert router1 == router2


@pytest.mark.parametrize("r1", [OnMessage()])
@pytest.mark.parametrize(
    "r2",
    [
        OnMessage(handlers=[object]),
        OnMessage(middleware=[object()]),
        OnMessage(guards=[object()]),
        OnMessage(handlers=[object], middleware=[object()]),
        OnMessage(handlers=[object], middleware=[object()], guards=[object()]),
    ]
)
def test_not_eq(r1, r2):
    assert r1 != r2



from __future__ import annotations

from typing import final

from yatbaf.types import WebhookInfo

from .abc import TelegramMethod


@final
class GetWebhookInfo(TelegramMethod[WebhookInfo]):
    """See :meth:`yatbaf.bot.Bot.get_webhook_info`"""



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from .abc import TelegramType
from .location import Location


@final
class Venue(TelegramType):
    """This object represents a venue.

    See: https://core.telegram.org/bots/api#venue
    """

    location: Location
    """Venue location. Can't be a live location."""

    title: str
    """Name of the venue."""

    address: str
    """Address of the venue."""

    foursquare_id: NoneStr = None
    """*Optional.* Foursquare identifier of the venue."""

    foursquare_type: NoneStr = None
    """*Optional.* Foursquare type of the venue."""

    google_place_id: NoneInt = None
    """*Optional.* Google Places identifier of the venue."""

    google_place_type: NoneInt = None
    """*Optional.* Google Places type of the venue."""



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneBool
from yatbaf.typing import NoneStr

from .abc import TelegramType
from .keyboard_button import KeyboardButton


@final
class ReplyKeyboardMarkup(TelegramType):
    """This object represents a custom keyboard with reply options.

    See: https://core.telegram.org/bots/api#replykeyboardmarkup
    """

    keyboard: list[list[KeyboardButton]]
    """List of button rows, each represented by an list of
    :class:`KyeboardButton <yatbaf.types.keyboard_button.KeyboardButton>`
    objects.
    """

    is_persistent: NoneBool = None
    """*Optional.* Requests clients to always show the keyboard when the regular
    keyboard is hidden. Defaults to ``False``, in which case the custom keyboard
    can be hidden and opened with a keyboard icon.
    """

    resize_keyboard: NoneBool = None
    """*Optional.* Requests clients to resize the keyboard vertically for
    optimal fit (e.g., make the keyboard smaller if there are just two rows of
    buttons). Defaults to ``False``, in which case the custom keyboard is always
    of the same height as the app's standard keyboard.
    """

    one_time_keyboard: NoneBool = None
    """*Optional.* Requests clients to hide the keyboard as soon as it's been
    used. The keyboard will still be available, but clients will automatically
    display the usual letter-keyboard in the chat - the user can press a special
    button in the input field to see the custom keyboard again. Defaults to
    ``False``.
    """

    input_field_placeholder: NoneStr = None
    """*Optional.* The placeholder to be shown in the input field when the
    keyboard is active; 1-64 characters.
    """

    selective: NoneBool = None
    """*Optional.* Use this parameter if you want to show the keyboard to
    specific users only. Targets: 1) users that are @mentioned in the text of
    the Message object; 2) if the bot's message is a reply (has
    ``reply_to_message_id``), sender of the original message.
    """



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.types import ChatAdministratorRights
    from yatbaf.typing import NoneBool


@final
class SetMyDefaultAdministratorRights(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.set_my_default_administrator_rights`"""

    rights: ChatAdministratorRights | None = None
    for_channels: NoneBool = None



import pytest

from yatbaf.utils import parse_command


@pytest.mark.parametrize(
    "text,result",
    [
        ("/foo", "foo"),
        ("/foo@bot", "foo"),
        ("/foo bar 123", "foo"),
        ("/foo@bot bar 123", "foo"),
        ("/FoO@bot bar 123", "foo"),
    ]
)
def test_parse_command(text, result):
    assert (r := parse_command(text)) == result, r


@pytest.mark.parametrize(
    "text", ["foo", "foo@bot", "foo bar 123", "foo@bot bar 123", "/"]
)
def test_parse_command_none(text):
    assert parse_command(text) is None



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class ChatShared(TelegramType):
    """This object contains information about the chat whose identifier was
    shared with the bot using a
    :class:`yatbaf.types.KeyboardButtonRequestChat` button.

    See: https://core.telegram.org/bots/api#chatshared
    """

    request_id: int
    """Identifier of the request."""

    chat_id: int
    """Identifier of the shared chat.

    .. note::

        The bot may not have access to the chat and could be unable to use
        this identifier, unless the chat is already known to the bot by some
        other means.
    """



from yatbaf.enums import UpdateType
from yatbaf.filters import Command
from yatbaf.filters import User
from yatbaf.handler import Handler


def test_match_empty_true(handler_func, message):
    handler = Handler(
        handler_func,
        update_type=UpdateType.MESSAGE,
    )
    assert handler._match(message)


def test_is_fallback(handler_func):
    handler = Handler(
        handler_func,
        update_type=UpdateType.MESSAGE,
        filters=[Command("foo")],
    )
    assert not handler._is_fallback()


def test_empty_is_fallback(handler_func):
    handler = Handler(
        handler_func,
        update_type=UpdateType.MESSAGE,
    )
    assert handler._is_fallback()


def test_match_true(handler_func, message):
    message.text = "/foo"
    handler = Handler(
        handler_func,
        update_type=UpdateType.MESSAGE,
        filters=[Command("foo")],
    )
    assert handler._match(message)


def test_match_true1(handler_func, message):
    message.text = "/foo"
    handler = Handler(
        handler_func,
        update_type=UpdateType.MESSAGE,
        filters=[
            Command("foo"),
            User(message.from_.id),
        ]
    )
    assert handler._match(message)


def test_match_false(handler_func, message):
    message.text = "/bar"
    handler = Handler(
        handler_func,
        update_type=UpdateType.MESSAGE,
        filters=[Command("foo")],
    )
    assert not handler._match(message)


def test_match_false1(handler_func, message):
    message.text = "/foo"
    handler = Handler(
        handler_func,
        update_type=UpdateType.MESSAGE,
        filters=[
            Command("bar"),
            User(1),
        ],
    )
    assert not handler._match(message)



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.typing import NoneBool


@final
class UnbanChatMember(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.unban_chat_member`"""

    chat_id: str | int
    user_id: str | int
    only_if_banned: NoneBool = None



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import BotDescription

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.typing import NoneStr


@final
class GetMyDescription(TelegramMethod[BotDescription]):
    """See :meth:`yatbaf.bot.Bot.get_my_description`"""

    language_code: NoneStr = None



import pytest

from yatbaf.enums import ChatType
from yatbaf.filters import Channel
from yatbaf.filters import Group

CHAT_ID = 123232


@pytest.fixture(autouse=True)
def _set_chat_id(chat):
    chat.id = CHAT_ID


@pytest.mark.parametrize("t", [ChatType.GROUP, ChatType.SUPERGROUP])
def test_filter_group(message, t):
    message.chat.type = t
    assert Group().check(message)


@pytest.mark.parametrize("t", [ChatType.GROUP, ChatType.SUPERGROUP])
def test_filter_group_ids(message, t):
    message.chat.type = t
    assert Group(CHAT_ID).check(message)
    assert not Group(123).check(message)


def test_filter_channel(message):
    message.chat.type = ChatType.CHANNEL
    assert Channel().check(message)


def test_filter_channel_ids(message):
    message.chat.type = ChatType.CHANNEL
    assert Channel(CHAT_ID).check(message)
    assert not Channel(123).check(message)



__all__ = [
    "ParseMode",
    "UpdateType",
    "ChatAction",
    "StickerType",
    "StickerFormat",
    "ThumbnailMimeType",
    "VideoMimeType",
    "DocumentMimeType",
    "PassportElement",
    "PollType",
    "ContentType",
    "IconColor",
    "Currency",
    "MaskPositionPoint",
    "MarkdownEntity",
]

from enum import Enum
from enum import IntEnum
from enum import StrEnum
from enum import auto


class ParseMode(StrEnum):
    """Message parse mode.

    See: https://core.telegram.org/bots/api#formatting-options
    """

    HTML = "HTML"
    MARKDOWN = "MarkdownV2"
    MARKDOWN_LEGACY = "Markdown"


class UpdateType(StrEnum):
    """Available update types.

    See :class:`~yatbaf.types.update.Update`
    """

    MESSAGE = auto()
    EDITED_MESSAGE = auto()
    CHANNEL_POST = auto()
    EDITED_CHANNEL_POST = auto()
    MESSAGE_REACTION = auto()
    MESSAGE_REACTION_COUNT = auto()
    INLINE_QUERY = auto()
    CHOSEN_INLINE_RESULT = auto()
    CALLBACK_QUERY = auto()
    SHIPPING_QUERY = auto()
    PRE_CHECKOUT_QUERY = auto()
    POLL = auto()
    POLL_ANSWER = auto()
    MY_CHAT_MEMBER = auto()
    CHAT_MEMBER = auto()
    CHAT_JOIN_REQUEST = auto()
    CHAT_BOOST = auto()
    REMOVED_CHAT_BOOST = auto()


class ChatAction(StrEnum):
    """Chat action.

    See :meth:`send_chat_action <yatbaf.bot.Bot.send_chat_action>`,
    https://core.telegram.org/bots/api#sendchataction
    """

    TYPING = auto()
    UPLOAD_PHOTO = auto()
    RECORD_VIDEO = auto()
    UPLOAD_VIDEO = auto()
    RECORD_VOICE = auto()
    UPLOAD_VOICE = auto()
    UPLOAD_DOCUMENT = auto()
    CHOOSE_STICKER = auto()
    FIND_LOCATION = auto()
    RECORD_VIDEO_NOTE = auto()
    UPLOAD_VIDEO_NOTE = auto()


class StickerType(StrEnum):
    REGULAR = auto()
    MASK = auto()
    CUSTOM_EMOJI = auto()


class StickerFormat(StrEnum):
    STATIC = auto()
    ANIMATED = auto()
    VIDEO = auto()


class ThumbnailMimeType(StrEnum):
    JPEG = "image/jpeg"
    GIF = "image/gif"
    MP4 = "video/mp4"


class VideoMimeType(StrEnum):
    HTML = "text/html"
    MP4 = "video/mp4"


class DocumentMimeType(StrEnum):
    PDF = "application/pdf"
    ZIP = "application/zip"


class PassportElement(StrEnum):
    ADDRESS = auto()
    PASSPORT = auto()
    DRIVER_LICENSE = auto()
    IDENTITY_CARD = auto()
    INTERNAL_PASSPORT = auto()
    UTILITY_BILL = auto()
    BANK_STATEMENT = auto()
    RENTAL_AGREEMENT = auto()
    PASSPORT_REGISTRATION = auto()
    TEMPORARY_REGISTRATION = auto()
    PERSONAL_DETAILS = auto()
    PHONE_NUMBER = auto()
    EMAIL = auto()


class PollType(StrEnum):
    """Poll type."""

    QUIZ = auto()
    REGULAR = auto()


# https://core.telegram.org/bots/api#messageentity
class MessageEntityType(StrEnum):
    MENTION = auto()
    HASHTAG = auto()
    CASHTAG = auto()
    BOT_COMMAND = auto()
    URL = auto()
    EMAIL = auto()
    PHONE_NUMBER = auto()
    BOLD = auto()
    ITALIC = auto()
    UNDERLINE = auto()
    STRIKETHROUGH = auto()
    SPOILER = auto()
    CODE = auto()
    PRE = auto()
    TEXT_LINK = auto()
    TEXT_MENTION = auto()
    CUSTOM_EMOJI = auto()


class ContentType(StrEnum):
    """Message content type.

    See: :class:`Message <yatbaf.types.message.Message>`
    """

    TEXT = auto()
    ANIMATION = auto()
    AUDIO = auto()
    DOCUMENT = auto()
    PHOTO = auto()
    STICKER = auto()
    VIDEO = auto()
    VIDEO_NOTE = auto()
    VOICE = auto()
    CONTACT = auto()
    DICE = auto()
    GAME = auto()
    POLL = auto()
    VENUE = auto()
    LOCATION = auto()
    PASSPORT_DATA = auto()
    INVOICE = auto()
    WEB_APP_DATA = auto()
    SUCCESSFUL_PAYMENT = auto()

    NEW_CHAT_MEMBER = auto()
    LEFT_CHAT_MEMBER = auto()
    NEW_CHAT_PHOTO = auto()
    DELETE_CHAT_PHOTO = auto()
    GROUP_CHAT_CREATED = auto()
    SUPERGROUP_CHAT_CREATED = auto()
    CHANNEL_CHAT_CREATED = auto()
    MESSAGE_AUTO_DELETE_TIMER_CHANGED = auto()
    MIGRATE_TO_CHAT_ID = auto()
    MIGRATE_FROM_CHAT_ID = auto()
    PINNED_MESSAGE = auto()
    USERS_SHARED = auto()
    CHAT_SHARED = auto()
    CONNECTED_WEBSITE = auto()
    WRITE_ACCESS_ALLOWED = auto()
    PROXIMITY_ALERT_TRIGGERED = auto()
    FORUM_TOPIC_CREATED = auto()
    FORUM_TOPIC_EDITED = auto()
    FORUM_TOPIC_CLOSED = auto()
    FORUM_TOPIC_REOPENED = auto()
    GENERAL_FORUM_TOPIC_HIDDEN = auto()
    GENERAL_FORUM_TOPIC_UNHIDDEN = auto()
    VIDEO_CHAT_SCHEDULED = auto()
    VIDEO_CHAT_STARTED = auto()
    VIDEO_CHAT_ENDED = auto()
    VIDEO_CHAT_PARTICIPANTS_INVITED = auto()


class ChatType(StrEnum):
    """Chat type.

    See :attr:`Chat.type <yatbaf.types.chat.Chat.type>`
    """

    SENDER = auto()
    PRIVATE = auto()
    GROUP = auto()
    SUPERGROUP = auto()
    CHANNEL = auto()


class IconColor(IntEnum):
    BLUE = 7322096  # 6FB9F0
    YELLOW = 16766590  # FFD67E
    PURPLE = 13338331  # CB86DB
    GREEN = 9367192  # 8EEE98
    PINK = 16749490  # FF93B2
    RED = 16478047  # FB6F5F


class Currency(StrEnum):
    """Supported currencies

    See: https://core.telegram.org/bots/payments#supported-currencies
    """

    UNITED_ARAB_EMIRATES_DIRHAM = "AED"
    AFGHAN_AFGHANI = "AFN"
    ALBANIAN_LEK = "ALL"
    ARMENIAN_DRAM = "AMD"
    ARGENTINE_PESO = "ARS"
    AUSTRALIAN_DOLLAR = "AUD"
    AZERBAIJANI_MANAT = "AZN"
    BOSNIA_HERZEGOVINA_CONVERTIBLE_MARK = "BAM"
    BANGLADESHI_TAKA = "BDT"
    BULGARIAN_LEV = "BGN"
    BRUNEI_DOLLAR = "BND"
    BOLIVIAN_BOLIVIANO = "BOB"
    BRAZILIAN_REAL = "BRL"
    BELARUSIAN_RUBLE = "BYN"
    CANADIAN_DOLLAR = "CAD"
    SWISS_FRANC = "CHF"
    CHILEAN_PESO = "CLP"
    CHINESE_RENMINBI_YUAN = "CNY"
    COLOMBIAN_PESO = "COP"
    COSTA_RICAN_COLON = "CRC"
    CZECH_KORUNA = "CZK"
    DANISH_KRONE = "DKK"
    DOMINICAN_PESO = "DOP"
    ALGERIAN_DINAR = "DZD"
    EGYPTIAN_POUND = "EGP"
    ETHIOPIAN_BIRR = "ETB"
    EURO = "EUR"
    BRITISH_POUND = "GBP"
    GEORGIAN_LARI = "GEL"
    GUATEMALAN_QUETZAL = "GTQ"
    HONG_KONG_DOLLAR = "HKD"
    HONDURAN_LEMPIRA = "HNL"
    CROATIAN_KUNA = "HRK"
    HUNGARIAN_FORINT = "HUF"
    INDONESIAN_RUPIAH = "IDR"
    ISRAELI_NEW_SHEQEL = "ILS"
    INDIAN_RUPEE = "INR"
    ICELANDIC_KRONA = "ISK"
    JAMAICAN_DOLLAR = "JMD"
    JAPANESE_YEN = "JPY"
    KENYAN_SHILLING = "KES"
    KYRGYZSTANI_SOM = "KGS"
    SOUTH_KOREAN_WON = "KRW"
    KAZAKHSTANI_TENGE = "KZT"
    LEBANESE_POUND = "LBP"
    SRI_LANKAN_RUPEE = "LKR"
    MOROCCAN_DIRHAM = "MAD"
    MOLDOVAN_LEU = "MDL"
    MONGOLIAN_TOGROG = "MNT"
    MAURITIAN_RUPEE = "MUR"
    MALDIVIAN_RUFIYAA = "MVR"
    MEXICAN_PESO = "MXN"
    MALAYSIAN_RINGGIT = "MYR"
    MOZAMBICAN_METICAL = "MZN"
    NIGERIAN_NAIRA = "NGN"
    NICARAGUAN_CORDOBA = "NIO"
    NORWEGIAN_KRONE = "NOK"
    NEPALESE_RUPEE = "NPR"
    NEW_ZEALAND_DOLLAR = "NZD"
    PANAMANIAN_BALBOA = "PAB"
    PERUVIAN_NUEVO_SOL = "PEN"
    PHILIPPINE_PESO = "PHP"
    PAKISTANI_RUPEE = "PKR"
    POLISH_ZLOTY = "PLN"
    PARAGUAYAN_GUARANI = "PYG"
    QATARI_RIYAL = "QAR"
    ROMANIAN_LEU = "RON"
    SERBIAN_DINAR = "RSD"
    RUSSIAN_RUBLE = "RUB"
    SAUDI_RIYAL = "SAR"
    SWEDISH_KRONA = "SEK"
    SINGAPORE_DOLLAR = "SGD"
    THAI_BAHT = "THB"
    TAJIKISTANI_SOMONI = "TJS"
    TURKISH_LIRA = "TRY"
    TRINIDAD_AND_TOBAGO_DOLLAR = "TTD"
    NEW_TAIWAN_DOLLAR = "TWD"
    TANZANIAN_SHILLING = "TZS"
    UKRAINIAN_HRYVNIA = "UAH"
    UGANDAN_SHILLING = "UGX"
    UNITED_STATES_DOLLAR = "USD"
    URUGUAYAN_PESO = "UYU"
    UZBEKISTANI_SOM = "UZS"
    VIETNAMESE_DONG = "VND"
    YEMENI_RIAL = "YER"
    SOUTH_AFRICAN_RAND = "ZAR"


class MaskPositionPoint(StrEnum):
    FOREHEAD = auto()
    EYES = auto()
    MOUTH = auto()
    CHIN = auto()


class BotEnvi(Enum):
    """See https://core.telegram.org/bots/features#dedicated-test-environment"""

    PROD = auto()
    TEST = auto()


class MarkdownEntity(StrEnum):
    """See https://core.telegram.org/bots/api#markdownv2-style"""

    TEXT = auto()
    """text"""
    FSTRING = auto()
    """text with f-string placeholder"""
    PRE = auto()
    """inline fixed-width code"""
    CODE = auto()
    """pre-formatted fixed-width code block"""
    LINK = auto()
    """inline url"""
    EMOJI = auto()
    """emoji"""


class AdminFlag(StrEnum):
    """See https://core.telegram.org/constructor/chatAdminRights#parameters"""

    CHANGE_INFO = auto()
    POST_MESSAGES = auto()
    EDIT_MESSAGES = auto()
    DELETE_MESSAGES = auto()
    RESTRICT_MEMBERS = auto()
    INVITE_USERS = auto()
    PIN_MESSAGES = auto()
    MANAGE_TOPICS = auto()
    PROMOTE_MEMBERS = auto()
    MANAGE_VIDEO_CHATS = auto()
    ANONYMOUS = auto()
    MANAGE_CHAT = auto()



from __future__ import annotations

from typing import final

from msgspec import field

from .abc import TelegramType
from .shipping_address import ShippingAddress
from .user import User


@final
class ShippingQuery(TelegramType):
    """This object contains information about an incoming shipping query.

    See: https://core.telegram.org/bots/api#shippingquery
    """

    id: str
    """Unique query identifier."""

    from_: User = field(name="from")
    """User who sent the query."""

    invoice_payload: str
    """Bot specified invoice payload."""

    shipping_address: ShippingAddress
    """User specified shipping address."""



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.types import MenuButton


@final
class SetChatMenuButton(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.set_chat_menu_button`"""

    chat_id: str | int | None = None
    menu_button: MenuButton | None = None



import asyncio
import sys
import unittest.mock as mock

import pytest
from httpx import ConnectError

from yatbaf import LongPolling
from yatbaf.exceptions import RequestTimeoutError

MODULE = "yatbaf.long_polling"
module_obj = sys.modules[MODULE]


@pytest.fixture
def asyncio_mock():
    asyncio_module = mock.Mock()
    asyncio_module.gather = mock.AsyncMock()
    asyncio_module.sleep = mock.AsyncMock()
    asyncio_module.wait_for = mock.AsyncMock()
    return asyncio_module


@pytest.fixture(autouse=True)
def __mock(monkeypatch, asyncio_mock):
    monkeypatch.setattr(module_obj, "asyncio", asyncio_mock)


@pytest.fixture
def bot_mock():
    bot = mock.AsyncMock()
    bot.tasks = mock.AsyncMock()
    bot.tasks.add = mock.Mock()
    bot.tasks._create_task = mock.Mock()
    bot._api_client = mock.Mock()
    bot._api_client.invoke = mock.AsyncMock()
    bot._api_client.close = mock.AsyncMock()
    return bot


@pytest.mark.asyncio
async def test_startup_running(asyncio_mock):
    polling = LongPolling(None)
    polling._running = True

    await polling._startup()
    asyncio_mock.get_running_loop.assert_not_called()


@pytest.mark.asyncio
async def test_startup(asyncio_mock):
    polling = LongPolling(None)
    asyncio_mock.get_running_loop.return_value = loop = mock.Mock()

    await polling._startup()
    polling._event.clear.assert_called_once()
    asyncio_mock.get_running_loop.assert_called_once()
    loop.add_signal_handler.assert_any_call(mock.ANY, polling.stop)
    assert polling._running


@pytest.mark.asyncio
async def test_startup_func(asyncio_mock, bot_mock):
    polling = LongPolling(
        bot_mock, on_startup=[
            mock.AsyncMock(),
            mock.AsyncMock(),
        ]
    )
    asyncio_mock.get_running_loop.return_value = mock.Mock()
    await polling._startup()
    for func in polling._on_startup:
        func.assert_awaited_once_with(bot_mock)


@pytest.mark.asyncio
async def test_startup_func_error(asyncio_mock, bot_mock):
    polling = LongPolling(
        bot_mock, on_startup=[
            mock.AsyncMock(side_effect=ValueError()),
        ]
    )
    with pytest.raises(ValueError):
        await polling._startup()
    assert not polling._running
    asyncio_mock.get_running_loop.assert_not_called()


@pytest.mark.asyncio
async def test_shutdown_not_running(bot_mock):
    polling = LongPolling(bot_mock)
    polling._running = False
    await polling._shutdown()
    bot_mock.shutdown.assert_not_awaited()
    bot_mock.tasks.wait_all.assert_not_awaited()


@pytest.mark.asyncio
async def test_shutdown_func(bot_mock):
    polling = LongPolling(
        bot_mock, on_shutdown=[
            mock.AsyncMock(),
            mock.AsyncMock(),
        ]
    )
    polling._running = True

    await polling._shutdown()
    bot_mock.tasks.wait_all.assert_awaited_once()
    bot_mock.shutdown.assert_awaited_once()
    for func in polling._on_shutdown:
        func.assert_awaited_once_with(bot_mock)


@pytest.mark.asyncio
async def test_shutdown_func_error(bot_mock):
    polling = LongPolling(
        bot_mock, on_shutdown=[
            mock.AsyncMock(side_effect=ValueError()),
        ]
    )
    polling._running = True
    with pytest.raises(ValueError):
        await polling._shutdown()
    bot_mock.shutdown.assert_awaited_once()


@pytest.mark.asyncio
async def test_get_updates_params(bot_mock):
    polling = LongPolling(bot_mock)
    polling._event.is_set.side_effect = [False, True]
    result = mock.Mock(result=[])
    bot_mock._api_client.invoke.return_value = result
    queue = mock.Mock()

    await polling._get_updates(queue)
    queue.put_nowait.assert_not_called()
    bot_mock._api_client.invoke.assert_awaited_once_with(
        polling._method,
        timeout=(polling._method.timeout + 5),
    )


@pytest.mark.asyncio
async def test_get_updates(bot_mock):
    polling = LongPolling(bot_mock)
    polling._event.is_set.side_effect = [False, False, True]
    result = mock.Mock(result=[])
    bot_mock._api_client.invoke.return_value = result
    queue = mock.Mock()

    await polling._get_updates(queue)
    queue.put_nowait.assert_not_called()
    bot_mock._api_client.invoke.assert_awaited()
    assert bot_mock._api_client.invoke.call_count == 2


@pytest.mark.asyncio
async def test_get_updates_queue(bot_mock, update):
    polling = LongPolling(bot_mock)
    polling._event.is_set.side_effect = [False, True]
    result = mock.Mock(result=[object(), update])
    bot_mock._api_client.invoke.return_value = result
    queue = mock.Mock()

    await polling._get_updates(queue)
    queue.put_nowait.assert_called_once_with(result.result)
    bot_mock._api_client.invoke.assert_awaited_once()
    assert polling._method.offset == update.update_id + 1
    assert bot_mock._api_client.invoke.call_count == 1


@pytest.mark.asyncio
async def test_get_updates_timeout_error(bot_mock, asyncio_mock):
    bot_mock._error_handler = error_handler = mock.AsyncMock()
    polling = LongPolling(bot_mock)
    polling._event.is_set.side_effect = [False, False, True]
    result = mock.Mock(result=[])
    bot_mock._api_client.invoke.side_effect = [
        result, RequestTimeoutError("", None), result
    ]
    queue = mock.Mock()

    await polling._get_updates(queue)
    error_handler.on_error.assert_not_awaited()
    asyncio_mock.sleep.assert_awaited_once_with(polling._timeout_delay)
    assert bot_mock._api_client.invoke.call_count == 2


@pytest.mark.asyncio
async def test_get_updates_connection_error(bot_mock, asyncio_mock):
    bot_mock._error_handler = error_handler = mock.AsyncMock()
    polling = LongPolling(bot_mock)
    polling._event.is_set.side_effect = [False, False, True]
    result = mock.Mock(result=[])
    bot_mock._api_client.invoke.side_effect = [result, ConnectError(""), result]
    queue = mock.Mock()

    await polling._get_updates(queue)
    error_handler.on_error.assert_not_awaited()
    asyncio_mock.sleep.assert_awaited_once_with(polling._connection_delay)
    assert bot_mock._api_client.invoke.call_count == 2


@pytest.mark.asyncio
async def test_get_updates_error(bot_mock):
    bot_mock._error_handler = error_handler = mock.AsyncMock()
    polling = LongPolling(bot_mock)
    polling._event = asyncio.Event()
    bot_mock._api_client.invoke.side_effect = exc = ValueError()
    queue = mock.Mock()
    await polling._get_updates(queue)
    error_handler.on_error.assert_awaited_once_with(exc)


@pytest.mark.asyncio
async def test_main_loop(monkeypatch, bot_mock, asyncio_mock, update):
    bot_mock.process_update = mock.Mock(
        return_value=(process_update_coro := object())
    )
    asyncio_mock.Queue.return_value = queue = mock.Mock()
    asyncio_mock.wait_for.return_value = [update]
    queue.get.return_value = get_coro = object()

    monkeypatch.setattr(
        LongPolling,
        "_get_updates",
        get_updates := mock.Mock(return_value=(updates_coro := object()))
    )
    asyncio_mock.create_task.side_effect = [updates_task := mock.Mock()]
    polling = LongPolling(bot_mock)
    polling._event.is_set.side_effect = [False, True]

    await polling._main_loop()

    get_updates.assert_called_once_with(queue)
    asyncio_mock.create_task.assert_any_call(updates_coro)
    bot_mock.tasks._create_task.assert_any_call(
        process_update_coro,
        name=f"update-task-{update.update_id}",
    )
    bot_mock.process_update.assert_called_once_with(update)
    asyncio_mock.wait_for.assert_awaited_once_with(get_coro, 0.2)
    bot_mock.tasks.kill.assert_awaited_once_with(updates_task)


@pytest.mark.asyncio
async def test_main_loop_timeout(monkeypatch, bot_mock, asyncio_mock):
    bot_mock.process_update = mock.Mock(return_value=object())
    asyncio_mock.Queue.return_value = mock.Mock()
    asyncio_mock.wait_for.side_effect = TimeoutError()

    monkeypatch.setattr(LongPolling, "_get_updates", mock.Mock())
    asyncio_mock.create_task.return_value = updates_task = mock.Mock()
    polling = LongPolling(bot_mock)
    polling._event.is_set.side_effect = [False, False, True]

    await polling._main_loop()

    assert asyncio_mock.create_task.call_count == 1
    assert asyncio_mock.wait_for.call_count == 2
    bot_mock.task._create_task.assert_not_called()
    bot_mock.process_update.assert_not_called()
    bot_mock.tasks.kill.assert_awaited_once_with(updates_task)


@pytest.mark.asyncio
async def test_run(monkeypatch):
    monkeypatch.setattr(LongPolling, "_startup", startup := mock.AsyncMock())
    monkeypatch.setattr(LongPolling, "_main_loop", loop := mock.AsyncMock())
    monkeypatch.setattr(LongPolling, "_shutdown", shutdown := mock.AsyncMock())

    polling = LongPolling(None)
    await polling._run()
    startup.assert_awaited_once()
    loop.assert_awaited_once()
    shutdown.assert_awaited_once()


@pytest.mark.asyncio
async def test_run_startup_error(monkeypatch):
    monkeypatch.setattr(
        LongPolling,
        "_startup",
        startup := mock.AsyncMock(side_effect=ValueError())
    )
    monkeypatch.setattr(LongPolling, "_main_loop", loop := mock.AsyncMock())
    monkeypatch.setattr(LongPolling, "_shutdown", shutdown := mock.AsyncMock())

    polling = LongPolling(None)
    await polling._run()
    assert not polling._running
    startup.assert_awaited_once()
    loop.assert_not_awaited()
    shutdown.assert_not_awaited()


def test_start(monkeypatch, asyncio_mock):
    monkeypatch.setattr(
        LongPolling,
        "_run",
        mock.Mock(return_value=(coro := mock.Mock())),
    )
    LongPolling(None).start()
    asyncio_mock.run.assert_called_once_with(coro)


def test_stop():
    polling = LongPolling(None)
    polling.stop()
    polling._event.set.assert_called_once()



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import final

from .abc import TelegramMethodWithMedia

if TYPE_CHECKING:
    from yatbaf.enums import StickerFormat
    from yatbaf.enums import StickerType
    from yatbaf.types import InputSticker
    from yatbaf.typing import NoneBool


@final
class CreateNewStickerSet(TelegramMethodWithMedia[bool]):
    """See :meth:`yatbaf.bot.Bot.create_new_sticker_set`"""

    user_id: int
    name: str
    title: str
    stickers: list[InputSticker]
    sticker_format: StickerFormat
    sticker_type: StickerType | None = None
    needs_repainting: NoneBool = None

    __meth_media_fields__: ClassVar[tuple[str, ...]] = ("stickers",)



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneStr

from .abc import TelegramType
from .shipping_address import ShippingAddress


@final
class OrderInfo(TelegramType):
    """This object represents information about an order.

    See: https://core.telegram.org/bots/api#orderinfo
    """

    name: NoneStr = None
    """*Optional.* User name."""

    phone_number: NoneStr = None
    """*Optional.* User's phone number."""

    email: NoneStr = None
    """*Optional.* User email."""

    shipping_address: ShippingAddress | None = None
    """*Optional.* User shipping address."""



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import final

from yatbaf.types import Message

from .abc import TelegramMethodWithMedia

if TYPE_CHECKING:
    from yatbaf.types import InputMediaAudio
    from yatbaf.types import InputMediaDocument
    from yatbaf.types import InputMediaPhoto
    from yatbaf.types import InputMediaVideo
    from yatbaf.types import ReplyParameters
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt


@final
class SendMediaGroup(TelegramMethodWithMedia[list[Message]]):
    """See :meth:`yatbaf.bot.Bot.send_media_group`"""

    chat_id: str | int
    # yapf: disable
    media: list[InputMediaAudio | InputMediaDocument | InputMediaPhoto | InputMediaVideo]  # noqa: E501
    # yapf: enable
    message_thread_id: NoneInt = None
    disable_notification: NoneBool = None
    protect_content: NoneBool = None
    reply_parameters: ReplyParameters | None = None

    __meth_media_fields__: ClassVar[tuple[str, ...]] = ("media",)



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from .abc import TelegramType


@final
class File(TelegramType):
    """This object represents a file ready to be downloaded.

    See: https://core.telegram.org/bots/api#file,
    :meth:`~yatbaf.bot.Bot.get_file`, :meth:`~yatbaf.bot.Bot.download_file`
    """

    file_id: str
    """Identifier for this file, which can be used to download or reuse the file."""  # noqa: E501

    file_unique_id: str
    """Unique identifier for this file, which is supposed to be the same over
    time and for different bots.

    .. warning::

        Can't be used to download or reuse the file.
    """

    file_size: NoneInt = None
    """*Optional.* File size in bytes."""

    file_path: NoneStr = None
    """*Optional.* File path."""



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneInt

from .abc import TelegramType


@final
class PhotoSize(TelegramType):
    """This object represents one size of a photo or a file / sticker thumbnail.

    See: https://core.telegram.org/bots/api#photosize
    """

    file_id: str
    """Identifier for this file, which can be used to download or reuse the file."""  # noqa: E501

    file_unique_id: str
    """Unique identifier for this file, which is supposed to be the same over
    time and for different bots.

    .. warning::

        Can't be used to download or reuse the file.
    """

    width: int
    """Photo width."""

    height: int
    """Photo height."""

    file_size: NoneInt = None
    """*Optional.* File size in bytes."""



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.types import InlineQueryResult
    from yatbaf.types import InlineQueryResultsButton
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt
    from yatbaf.typing import NoneStr


@final
class AnswerInlineQuery(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.answer_inline_query`"""

    inline_query_id: str
    results: list[InlineQueryResult]
    cache_time: NoneInt = None
    is_personal: NoneBool = None
    next_offset: NoneStr = None
    button: InlineQueryResultsButton | None = None



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneStr

from .abc import TelegramType
from .order_info import OrderInfo


@final
class SuccessfulPayment(TelegramType, kw_only=True):
    """This object contains basic information about a successful payment.

    See: https://core.telegram.org/bots/api#successfulpayment
    """

    currency: str
    """Three-letter ISO 4217 currency code."""

    total_amount: int
    """Total price in the smallest units of the currency.

    .. important::

        Integer, not float/double.
    """

    invoice_payload: str
    """Bot specified invoice payload."""

    shipping_option_id: NoneStr = None
    """*Optional.* Identifier of the shipping option chosen by the user."""

    order_info: OrderInfo | None = None
    """*Optional.* Order information provided by the user."""

    telegram_payment_charge_id: str
    """Telegram payment identifier."""

    provider_payment_charge_id: str
    """Provider payment identifier."""



__all__ = (
    "Channel",
    "Chat",
    "Group",
    "Command",
    "Content",
    "And",
    "Not",
    "Or",
    "Text",
    "User",
)

from .chat import Channel
from .chat import Chat
from .chat import Group
from .command import Command
from .content import Content
from .logical import And
from .logical import Not
from .logical import Or
from .text import Text
from .user import User



from __future__ import annotations

__all__ = (
    "PassportElementError",
    "PassportElementErrorDataField",
    "PassportElementErrorFile",
    "PassportElementErrorFiles",
    "PassportElementErrorFrontSide",
    "PassportElementErrorReverseSide",
    "PassportElementErrorSelfie",
    "PassportElementErrorTranslationFile",
    "PassportElementErrorTranslationFiles",
    "PassportElementErrorUnspecified",
)

from typing import TYPE_CHECKING
from typing import Literal
from typing import TypeAlias
from typing import final

from msgspec import field

if TYPE_CHECKING:
    from yatbaf.enums import PassportElement

from .abc import TelegramType


@final
class PassportElementErrorFile(TelegramType):
    """Represents an issue with a document scan. The error is considered
    resolved when the file with the document scan changes.

    See: https://core.telegram.org/bots/api#passportelementerrorfile
    """

    type: PassportElement
    """The section of the user's Telegram Passport which has the issue."""

    file_hash: str
    """Base64-encoded file hash."""

    message: str
    """Error message"""

    source: Literal["file"] = field(default_factory=lambda: "file")
    """Error source, must be `file`."""


@final
class PassportElementErrorFiles(TelegramType):
    """Represents an issue with a list of scans. The error is considered
    resolved when the list of files containing the scans changes.

    See: https://core.telegram.org/bots/api#passportelementerrorfiles
    """

    type: PassportElement
    """The section of the user's Telegram Passport which has the issue."""

    file_hashes: list[str]
    """List of base64-encoded file hashes."""

    message: str
    """Error message."""

    source: Literal["files"] = field(default_factory=lambda: "files")
    """Error source, must be `files`."""


@final
class PassportElementErrorSelfie(TelegramType):
    """Represents an issue with the selfie with a document. The error is
    considered resolved when the file with the selfie changes.

    # https://core.telegram.org/bots/api#passportelementerrorselfie
    """

    type: PassportElement
    """The section of the user's Telegram Passport which has the issue."""

    file_hash: str
    """Base64-encoded hash of the file with the selfie."""

    message: str
    """Error message."""

    source: Literal["selfie"] = field(default_factory=lambda: "selfie")
    """Error source, must be `selfie`."""


@final
class PassportElementErrorFrontSide(TelegramType):
    """Represents an issue with the front side of a document. The error is
    considered resolved when the file with the front side of the document
    changes.

    See: https://core.telegram.org/bots/api#passportelementerrorfrontside
    """

    type: PassportElement
    """The section of the user's Telegram Passport which has the issue."""

    file_hash: str
    """Base64-encoded hash of the file with the front side of the document."""

    message: str
    """Error message."""

    source: Literal["front_side"] = field(default_factory=lambda: "front_side")
    """Error source, must be `front_side`."""


@final
class PassportElementErrorUnspecified(TelegramType):
    """Represents an issue in an unspecified place. The error is considered
    resolved when new data is added.

    See: https://core.telegram.org/bots/api#passportelementerrorunspecified
    """

    type: PassportElement
    """Type of element of the user's Telegram Passport which has the issue."""

    element_hash: str
    """Base64-encoded element hash."""

    message: str
    """Error message."""

    source: Literal["unspecified"] = field(
        default_factory=lambda: "unspecified"
    )
    """Error source, must be `unspecified`."""


@final
class PassportElementErrorReverseSide(TelegramType):
    """Represents an issue with the reverse side of a document. The error is
    considered resolved when the file with reverse side of the document changes.

    See: https://core.telegram.org/bots/api#passportelementerrorreverseside
    """

    type: PassportElement
    """The section of the user's Telegram Passport which has the issue."""

    file_hash: str
    """Base64-encoded hash of the file with the reverse side of the document."""

    message: str
    """Error message."""

    source: Literal["reverse_side"] = field(
        default_factory=lambda: "reverse_side"
    )
    """Error source, must be `reverse_side`."""


@final
class PassportElementErrorTranslationFile(TelegramType):
    """Represents an issue with one of the files that constitute the translation
    of a document. The error is considered resolved when the file changes.

    See: https://core.telegram.org/bots/api#passportelementerrortranslationfile
    """

    type: PassportElement
    """ype of element of the user's Telegram Passport which has the issue."""

    file_hash: str
    """Base64-encoded file hash."""

    message: str
    """Error message."""

    source: Literal["translation_file"] = field(
        default_factory=lambda: "translation_file"
    )
    """Error source, must be `translation_file`."""


@final
class PassportElementErrorTranslationFiles(TelegramType):
    """Represents an issue with the translated version of a document. The error
    is considered resolved when a file with the document translation change.

    See: https://core.telegram.org/bots/api#passportelementerrortranslationfiles
    """

    type: PassportElement
    """Type of element of the user's Telegram Passport which has the issue."""

    file_hashes: list[str]
    """List of base64-encoded file hashes."""

    message: str
    """Error message."""

    source: Literal["translation_files"] = field(
        default_factory=lambda: "translation_files"
    )
    """Error source, must be `translation_files`."""


@final
class PassportElementErrorDataField(TelegramType):
    """Represents an issue in one of the data fields that was provided by the
    user. The error is considered resolved when the field's value changes.

    See: https://core.telegram.org/bots/api#passportelementerrordatafield
    """

    type: PassportElement
    """The section of the user's Telegram Passport which has the error."""

    field_name: str
    """Name of the data field which has the error."""

    data_hash: str
    """Base64-encoded data hash."""

    message: str
    """Error message"""

    source: Literal["data"] = field(default_factory=lambda: "data")
    """Error source, must be `data`."""


# https://core.telegram.org/bots/api#passportelementerror
# Represents an error in the Telegram Passport element which was submitted that
# should be resolved by the user.
PassportElementError: TypeAlias = (
    "PassportElementErrorDataField "
    "| PassportElementErrorFile "
    "| PassportElementErrorFiles "
    "| PassportElementErrorFrontSide "
    "| PassportElementErrorReverseSide "
    "| PassportElementErrorSelfie "
    "| PassportElementErrorTranslationFile "
    "| PassportElementErrorTranslationFiles "
    "| PassportElementErrorUnspecified"
)



from __future__ import annotations

from typing import ClassVar
from typing import Literal
from typing import TypeAlias
from typing import final

from yatbaf.typing import NoneStr

from .abc import TelegramType
from .chat import Chat
from .user import User


@final
class MessageOriginUser(TelegramType, tag="user"):
    """The message was originally sent by a known user.

    See: https://core.telegram.org/bots/api#messageoriginuser
    """

    date: int
    """Date the message was sent originally in Unix time."""

    sender_user: User
    """User that sent the message originally."""

    type: ClassVar[Literal["user"]] = "user"
    """Type of the message origin, always `user`"""


@final
class MessageOriginHiddenUser(TelegramType, tag="hidden_user"):
    """The message was originally sent by an unknown user.

    See: https://core.telegram.org/bots/api#messageoriginhiddenuser
    """

    date: int
    """Date the message was sent originally in Unix time."""

    sender_user_name: str
    """Name of the user that sent the message originally."""

    type: ClassVar[Literal["hidden_user"]] = "hidden_user"
    """Type of the message origin, always `hidden_user`"""


@final
class MessageOriginChat(TelegramType, tag="chat"):
    """The message was originally sent on behalf of a chat to a group chat.

    See: https://core.telegram.org/bots/api#messageoriginchat
    """

    date: int
    """Date the message was sent originally in Unix time."""

    sender_chat: Chat
    """Chat that sent the message originally."""

    author_signature: NoneStr = None
    """*Optional.* For messages originally sent by an anonymous chat
    administrator, original message author signature.
    """

    type: ClassVar[Literal["chat"]] = "chat"
    """Type of the message origin, always `chat`"""


@final
class MessageOriginChannel(TelegramType, tag="channel"):
    """The message was originally sent to a channel chat.

    See: https://core.telegram.org/bots/api#messageoriginchannel
    """

    date: int
    """Date the message was sent originally in Unix time."""

    chat: Chat
    """Channel chat to which the message was originally sent."""

    message_id: int
    """Unique message identifier inside the chat."""

    author_signature: NoneStr = None
    """*Optional.* Signature of the original post author."""

    type: ClassVar[Literal["channel"]] = "channel"
    """Type of the message origin, always `channel`"""


MessageOrigin: TypeAlias = (
    MessageOriginUser
    | MessageOriginHiddenUser
    | MessageOriginChat
    | MessageOriginChannel
)
"""This object describes the origin of a message.

See: https://core.telegram.org/bots/api#messageorigin
"""



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import UserProfilePhotos

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.typing import NoneInt


@final
class GetUserProfilePhotos(TelegramMethod[UserProfilePhotos]):
    """See :meth:`yatbaf.bot.Bot.get_user_profile_photos`"""

    user_id: int
    offset: NoneInt = None
    limit: NoneInt = None



import unittest.mock as mock

import pytest

from yatbaf.filters import Command
from yatbaf.handler import Handler
from yatbaf.middleware import Middleware
from yatbaf.router import OnMessage


@pytest.mark.asyncio
async def test_resolve_no_handlers(router, update_info):
    router._on_registration()
    assert not await router._resolve(update_info)


@pytest.mark.asyncio
async def test_resolve_catch_all(router, update_info, handler_func):
    router.add_handler(handler_func)
    router._on_registration()
    assert await router._resolve(update_info)
    handler_func.assert_awaited_once()


@pytest.mark.asyncio
async def test_resolve_none(router, update_info, handler_func):
    filter = Command("foo")
    router.add_handler(handler_func, filters=[filter])
    router._on_registration()
    assert not await router._resolve(update_info)
    handler_func.assert_not_awaited()


@pytest.mark.asyncio
async def test_resolve_nested_vert(update_info, handler_func):

    def find_router(router):
        r = router
        while r._routers:
            r = r._routers[-1]
        return r

    router = OnMessage()
    for _ in range(5):
        find_router(router).add_router(OnMessage())

    filter = Command("foo")
    update_info.content.text = "/foo"
    find_router(router).add_handler(handler_func, filters=[filter])
    router._on_registration()
    assert await router._resolve(update_info)
    handler_func.assert_awaited_once()


@pytest.mark.asyncio
async def test_resolve_nested_horiz(update_info, handler_func):
    router = OnMessage(
        routers=[
            OnMessage(
                handlers=[
                    Handler(
                        mock.AsyncMock(),
                        filters=[Command(f"bar{i}")],
                        update_type="message",
                    ),
                ]
            ) for i in range(5)
        ],
    )
    router.add_router(
        OnMessage(
            handlers=[
                Handler(
                    handler_func,
                    filters=[Command("foo")],
                    update_type="message",
                ),
            ]
        )
    )
    router._on_registration()
    update_info.content.text = "/foo"
    assert await router._resolve(update_info)
    handler_func.assert_awaited_once()


@pytest.mark.asyncio
async def test_guard_false(handler_func, router, update_info, asyncdef):
    router.add_guard(asyncdef(False))
    router.add_handler(handler_func)
    router._on_registration()
    assert not await router._resolve(update_info)
    handler_func.assert_not_awaited()


@pytest.mark.asyncio
async def test_guard_true(handler_func, router, update_info, asyncdef):
    router.add_guard(asyncdef(True))
    router.add_handler(handler_func)
    router._on_registration()
    assert await router._resolve(update_info)
    handler_func.assert_awaited_once()


@pytest.mark.asyncio
async def test_parent_guard_true(handler_func, update_info, asyncdef):
    router = OnMessage()
    router.add_guard(asyncdef(True))
    router.add_handler(handler_func)

    router1 = OnMessage()
    router1.add_handler(asyncdef())

    router.add_router(router1)
    router._on_registration()
    assert await router._resolve(update_info)
    handler_func.assert_awaited_once()


@pytest.mark.asyncio
async def test_parent_guard_false(handler_func, update_info, asyncdef):
    router = OnMessage()
    router.add_guard(asyncdef(False))
    router.add_handler(asyncdef())

    router1 = OnMessage()
    router1.add_handler(handler_func)

    router.add_router(router1)
    router._on_registration()
    assert await router._resolve(update_info)
    handler_func.assert_awaited_once()


@pytest.mark.asyncio
async def test_guard_false_both(handler_func, update_info, asyncdef):
    router = OnMessage()
    router.add_guard(asyncdef(False))
    router.add_handler(asyncdef())

    router1 = OnMessage()
    router1.add_guard(asyncdef(False))
    router1.add_handler(handler_func)

    router.add_router(router1)
    router._on_registration()
    assert not await router._resolve(update_info)
    handler_func.assert_not_awaited()


@pytest.mark.asyncio
async def test_guard_skip_nested(handler_func, update_info, asyncdef):
    router = OnMessage(skip_with_nested=True)
    router.add_guard(asyncdef(False))

    router1 = OnMessage()
    router1.add_guard(asyncdef(True))
    router1.add_handler(handler_func)

    router.add_router(router1)
    assert not await router._resolve(update_info)
    handler_func.assert_not_awaited()


# yapf: disable
def middleware_factory(mark):
    def middleware(handler):
        async def wrapper(update):
            mark()
            await handler(update)
        return wrapper
    return middleware
# yapf: enable


@pytest.mark.asyncio
async def test_resolve_wrap_router_middlewares(
    router, update_info, handler_func
):
    m = mock.Mock()
    router.add_handler(handler_func)
    for _ in range(5):
        router.add_middleware(
            Middleware(
                middleware_factory(m),
                is_handler=True,
            )
        )
    assert len(router._middleware) == 5

    router._on_registration()
    assert await router._resolve(update_info)
    handler_func.assert_awaited_once_with(update_info.content)
    assert m.call_count == 5


@pytest.mark.asyncio
async def test_resolve_wrap_handler_middlewares(
    router, handler_func, update_info
):
    m = mock.Mock()
    router.add_handler(
        handler_func,
        middleware=[middleware_factory(m) for _ in range(5)],
    )
    router._on_registration()
    assert await router._resolve(update_info)
    assert m.call_count == 5
    handler_func.assert_awaited_once_with(update_info.content)


@pytest.mark.asyncio
async def test_resolve_wrap_middlewares(router, update_info, handler_func):
    m = mock.Mock()
    router.add_handler(
        handler_func,
        middleware=[middleware_factory(m) for _ in range(5)],
    )
    for _ in range(5):
        router.add_middleware(
            Middleware(
                middleware_factory(m),
                is_handler=True,
            )
        )
    router._on_registration()
    assert await router._resolve(update_info)
    assert m.call_count == 10
    handler_func.assert_awaited_once_with(update_info.content)



from __future__ import annotations

from typing import final

from yatbaf.types import User

from .abc import TelegramMethod


@final
class GetMe(TelegramMethod[User]):
    """See :meth:`yatbaf.bot.Bot.get_me`"""



from __future__ import annotations

__all__ = (
    "ChatBoostSource",
    "ChatBoostSourcePremium",
    "ChatBoostSourceGiftCode",
    "ChatBoostSourceGiveaway",
)

from typing import ClassVar
from typing import Literal
from typing import TypeAlias
from typing import final

from yatbaf.typing import NoneBool

from .abc import TelegramType
from .user import User


@final
class ChatBoostSourcePremium(TelegramType, tag="premium"):
    """The boost was obtained by subscribing to Telegram Premium or by gifting
    a Telegram Premium subscription to another user.

    See: https://core.telegram.org/bots/api#chatboostsourcepremium
    """

    user: User
    """User that boosted the chat."""

    source: ClassVar[Literal["premium"]] = "premium"
    """Source of the boost, always *premium*."""


@final
class ChatBoostSourceGiftCode(TelegramType, tag="gift_code"):
    """The boost was obtained by the creation of Telegram Premium gift codes to
    boost a chat. Each such code boosts the chat 4 times for the duration of the
    corresponding Telegram Premium subscription.

    See: https://core.telegram.org/bots/api#chatboostsourcegiftcode
    """

    user: User
    """User for which the gift code was created."""

    source: ClassVar[Literal["gift_code"]] = "gift_code"
    """Source of the boost, always *gift_code*."""


@final
class ChatBoostSourceGiveaway(TelegramType, tag="giveaway"):
    """The boost was obtained by the creation of a Telegram Premium giveaway.
    This boosts the chat 4 times for the duration of the corresponding Telegram
    Premium subscription.

    See: https://core.telegram.org/bots/api#chatboostsourcegiveaway
    """

    giveaway_message_id: int
    """Identifier of a message in the chat with the giveaway.

    .. note::

        The message could have been deleted already. May be 0 if the message
        isn't sent yet.
    """

    user: User | None = None
    """*Optional.* User that won the prize in the giveaway if any."""

    is_unclaimed: NoneBool = None
    """*Optional.* ``True``, if the giveaway was completed, but there was no
    user to win the prize.
    """

    source: ClassVar[Literal["giveaway"]] = "giveaway"
    """Source of the boost, always *giveaway*."""


ChatBoostSource: TypeAlias = (
    ChatBoostSourcePremium
    | ChatBoostSourceGiftCode
    | ChatBoostSourceGiveaway
)
"""This object describes the source of a chat boost.

See: https://core.telegram.org/bots/api#chatboostsource
"""



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneBool

from .abc import TelegramType
from .animation import Animation
from .audio import Audio
from .chat import Chat
from .contact import Contact
from .dice import Dice
from .document import Document
from .game import Game
from .giveaway import Giveaway
from .giveaway_winners import GiveawayWinners
from .invoice import Invoice
from .link_preview_options import LinkPreviewOptions
from .location import Location
from .message_origin import MessageOrigin
from .photo_size import PhotoSize
from .poll import Poll
from .sticker import Sticker
from .story import Story
from .venue import Venue
from .video import Video
from .video_note import VideoNote
from .voice import Voice


@final
class ExternalReplyInfo(TelegramType):
    """This object contains information about a message that is being replied
    to, which may come from another chat or forum topic.

    See: https://core.telegram.org/bots/api#externalreplyinfo
    """

    origin: MessageOrigin
    """Origin of the message replied to by the given message."""

    chat: Chat | None = None
    """*Optional.* Chat the original message belongs to. Available only if the
    chat is a supergroup or a channel."""

    message_id: int | None = None
    """*Optional.* Unique message identifier inside the original chat. Available
    only if the original chat is a supergroup or a channel.
    """

    link_preview_options: LinkPreviewOptions | None = None
    """*Optional.* Options used for link preview generation for the original
    message, if it is a text message.
    """

    animation: Animation | None = None
    """*Optional.* Message is an animation, information about the animation."""

    audio: Audio | None = None
    """*Optional.* Message is an audio file, information about the file."""

    document: Document | None = None
    """*Optional.* Message is a general file, information about the file."""

    photo: list[PhotoSize] | None = None
    """*Optional.* Message is a photo, available sizes of the photo."""

    sticker: Sticker | None = None
    """*Optional.* Message is a sticker, information about the sticker."""

    story: Story | None = None
    """*Optional.* Message is a forwarded story."""

    video: Video | None = None
    """*Optional.* Message is a video, information about the video."""

    video_note: VideoNote | None = None
    """*Optional.* Message is a video note, information about the video message."""  # noqa: E501

    voice: Voice | None = None
    """*Optional.* Message is a voice message, information about the file."""

    has_media_spoiler: NoneBool = None
    """*Optional.* True, if the message media is covered by a spoiler animation."""  # noqa: E501

    contact: Contact | None = None
    """*Optional.* Message is a shared contact, information about the contact."""  # noqa: E501

    dice: Dice | None = None
    """*Optional.* Message is a dice with random value."""

    game: Game | None = None
    """*Optional.* Message is a game, information about the game."""

    giveaway: Giveaway | None = None
    """*Optional.* Message is a scheduled giveaway, information about the giveaway."""  # noqa: E501

    giveaway_winners: GiveawayWinners | None = None
    """*Optional.* A giveaway with public winners was completed."""

    invoice: Invoice | None = None
    """*Optional.* Message is an invoice for a payment, information about the invoice."""  # noqa: E501

    location: Location | None = None
    """*Optional.* Message is a shared location, information about the location."""  # noqa: E501

    poll: Poll | None = None
    """*Optional.* Message is a native poll, information about the poll."""

    venue: Venue | None = None
    """*Optional.* Message is a venue, information about the venue."""



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.enums import UpdateType
from yatbaf.types import Update

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.typing import NoneInt


@final
class GetUpdates(TelegramMethod[list[Update]]):
    """See :meth:`yatbaf.bot.Bot.get_updates`"""

    offset: NoneInt = None
    limit: NoneInt = None
    timeout: float | None = None
    allowed_updates: list[UpdateType] | None = None



import unittest.mock as mock

from yatbaf.enums import UpdateType
from yatbaf.filters import Command
from yatbaf.handler import Handler


def test_empty_router(router, message):
    assert router._find_handler(message) is None


def test_handler_no_filters(router, message, handler):
    router.add_handler(handler)
    assert router._find_handler(message) is handler


def test_fallback_handler(router, message):
    fallback = Handler(object(), update_type=UpdateType.MESSAGE)
    router.add_handler(fallback)
    router.add_handler(object(), filters=[Command("foo")])
    message.text = "/bar"
    assert router._find_handler(message) is fallback


def test_filter(router, message):
    router._handlers = [
        Handler(
            object(),
            filters=[Command("ewq")],
            update_type=UpdateType.MESSAGE,
        ),
        handler := Handler(
            object(),
            filters=[Command("foo")],
            update_type=UpdateType.MESSAGE,
        ),
        handler_mock := mock.Mock(),
    ]
    message.text = "/foo"
    assert router._find_handler(message) is handler
    handler_mock._match.assert_not_called()



from __future__ import annotations

from typing import final

from .abc import TelegramMethod


@final
class SetChatDescription(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.set_chat_description`"""

    chat_id: str | int
    description: str | None = None



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneBool
from yatbaf.typing import NoneStr

from .abc import TelegramType


@final
class WriteAccessAllowed(TelegramType):
    """This object represents a service message about a user allowing a bot to
    write messages after adding the bot to the attachment menu or launching a
    Web App from a link.

    See: https://core.telegram.org/bots/api#writeaccessallowed
    """

    from_request: NoneBool = None
    """*Optional.* ``True``, if the access was granted after the user accepted
    an explicit request from a Web App sent by the method requestWriteAccess.
    """

    web_app_name: NoneStr = None
    """*Optional.* Name of the Web App which was launched from a link."""

    from_attachment_menu: NoneBool = None
    """*Optional.* ``True``, if the access was granted when the bot was added
    to the attachment or side menu.
    """



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.typing import NoneBool
from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from .abc import TelegramType

if TYPE_CHECKING:
    from yatbaf.types import UserProfilePhotos


@final
class User(TelegramType):
    """This object represents a Telegram user or bot.

    See: https://core.telegram.org/bots/api#user
    """

    id: int
    """Unique identifier for this user or bot."""

    is_bot: bool
    """``True``, if this user is a bot."""

    first_name: str
    """User's or bot's first name."""

    last_name: NoneStr = None
    """*Optional.* User's or bot's last name."""

    username: NoneStr = None
    """*Optional.* User's or bot's username."""

    language_code: NoneStr = None
    """*Optional.* IETF language tag of the user's language."""

    is_premium: NoneBool = None
    """*Optional.* ``True``, if this user is a Telegram Premium user."""

    added_to_attachment_menu: NoneBool = None
    """
    *Optional.* ``True``, if this user added the bot to the attachment menu.
    """

    can_join_groups: NoneBool = None
    """*Optional.* ``True``, if the bot can be invited to groups.

    .. note::

        Returned only in :meth:`get_me <yatbaf.bot.Bot.get_me>`.
    """

    can_read_all_group_messages: NoneBool = None
    """*Optional.* ``True``, if privacy mode is disabled for the bot.

    .. note::

        Returned only in :meth:`get_me <yatbaf.bot.Bot.get_me>`.
    """

    supports_inline_queries: NoneBool = None
    """*Optional.* ``True``, if the bot supports inline queries.

    .. note::

        Returned only in :meth:`get_me <yatbaf.bot.Bot.get_me>`.
    """

    async def get_photos(
        self,
        offset: NoneInt = None,
        limit: NoneInt = None,
    ) -> UserProfilePhotos:
        """Get profile photos.

        See: :meth:`get_user_profile_photos <yatbaf.bot.Bot.get_user_profile_photos>`
        """  # noqa: E501

        return await self.bot.get_user_profile_photos(
            user_id=self.id,
            offset=offset,
            limit=limit,
        )



from __future__ import annotations

__all__ = ["Chat"]

from typing import TYPE_CHECKING
from typing import final

from yatbaf.enums import ChatType
from yatbaf.typing import NoneBool
from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from ..types import message
from .abc import TelegramType
from .bot_command_scope import BotCommandScopeChat
from .bot_command_scope import BotCommandScopeChatAdministrators
from .bot_command_scope import BotCommandScopeChatMember
from .chat_location import ChatLocation
from .chat_permissions import ChatPermissions
from .chat_photo import ChatPhoto
from .reaction_type import ReactionType

if TYPE_CHECKING:
    from yatbaf.types import BotCommand
    from yatbaf.types import ChatInviteLink
    from yatbaf.types import ChatMember
    from yatbaf.typing import InputFile


@final
class Chat(TelegramType):
    """This object represents a chat.

    See: https://core.telegram.org/bots/api#chat
    """

    id: int
    """Unique identifier for this chat."""

    type: ChatType
    """Type of chat (see :class:`yatbaf.enums.ChatType`)."""

    title: NoneStr = None
    """*Optional.* Title, for supergroups, channels and group chats."""

    username: NoneStr = None
    """*Optional.* Username, for private chats, supergroups and channels if available."""  # noqa: E501

    first_name: NoneStr = None
    """*Optional.* First name of the other party in a private chat."""

    last_name: NoneStr = None
    """*Optional.* Last name of the other party in a private chat."""

    is_forum: NoneBool = None
    """*Optional.* ``True``, if the supergroup chat is a forum (has topics enabled)."""  # noqa: E501

    photo: ChatPhoto | None = None
    """*Optional.* Chat photo. Returned only in :meth:`~yatbaf.bot.Bot.get_chat`."""  # noqa: E501

    active_usernames: list[str] | None = None
    """*Optional.* If non-empty, the list of all active chat usernames;
    for private chats, supergroups and channels. Returned only in
    :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    available_reactions: list[ReactionType] | None = None
    """*Optional.* List of available reactions allowed in the chat. If omitted,
    then all emoji reactions are allowed. Returned only in
    :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    accent_color_id: NoneInt = None
    """*Optional.* Identifier of the accent color for the chat name and
    backgrounds of the chat photo, reply header, and link preview. See `accent colors`_
    for more details. Returned only in :meth:`~yatbaf.bot.Bot.get_chat`. Always
    returned in :meth:`~yatbaf.bot.Bot.get_chat`.

    .. _accent colors: https://core.telegram.org/bots/api#accent-colors
    """  # noqa: E501

    background_custom_emoji_id: NoneStr = None
    """*Optional.* Custom emoji identifier of emoji chosen by the chat for the
    reply header and link preview background. Returned only in
    :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    profile_accent_color_id: NoneInt = None
    """*Optional.* Identifier of the accent color for the chat's profile
    background. See `profile accent colors`_ for more details. Returned only in
    :meth:`~yatbaf.bot.Bot.get_chat`.

    .. _profile accent colors: https://core.telegram.org/bots/api#profile-accent-colors
    """  # noqa: E501

    profile_background_custom_emoji_id: NoneStr = None
    """*Optional.* Custom emoji identifier of the emoji chosen by the chat for
    its profile background. Returned only in :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    emoji_status_custom_emoji_id: NoneStr = None
    """*Optional.* Custom emoji identifier of emoji status of the other party
    in a private chat. Returned only in :meth:`~yatbaf.bot.Bot.get_chat`.
    """  # noqa: E501

    emoji_status_expiration_date: NoneInt = None
    """*Optional.* Expiration date of the emoji status of the other party in a
    private chat, if any. Returned only in :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    bio: NoneStr = None
    """*Optional.* Bio of the other party in a private chat. Returned only in
    :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    has_private_forwards: NoneBool = None
    """*Optional.* ``True``, if privacy settings of the other party in the
    private chat allows to use tg://user?id=<user_id> links only in chats with
    the user. Returned only in :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    has_restricted_voice_and_video_messages: NoneBool = None
    """*Optional.* ``True``, if the privacy settings of the other party restrict
    sending voice and video note messages in the private chat. Returned only in
    :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    join_to_send_messages: NoneBool = None
    """*Optional.* True, if users need to join the supergroup before they can
    send messages. Returned only in :meth:`~yatbaf.bot.Bot.get_chat`.
    """  # noqa: E501

    join_by_request: NoneBool = None
    """*Optional.* True, if all users directly joining the supergroup need to
    be approved by supergroup administrators. Returned only in
    :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    description: NoneStr = None
    """*Optional.* Description, for groups, supergroups and channel chats.
    Returned only in :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    invite_link: NoneStr = None
    """*Optional.* Primary invite link, for groups, supergroups and channel
    chats. Returned only in :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    pinned_message: message.Message | None = None
    """*Optional.* The most recent pinned message (by sending date).
    Returned only in :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    permissions: ChatPermissions | None = None
    """*Optional.* Default chat member permissions, for groups and supergroups.
    Returned only in :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    slow_mode_delay: NoneInt = None
    """*Optional.* For supergroups, the minimum allowed delay between
    consecutive messages sent by each unpriviledged user; in seconds. Returned
    only in :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    message_auto_delete_time: NoneInt = None
    """*Optional.* The time after which all messages sent to the chat will be
    automatically deleted; in seconds. Returned only in
    :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    has_aggressive_anti_spam_enabled: NoneBool = None
    """*Optional.* ``True``, if aggressive anti-spam checks are enabled in the
    supergroup. The field is only available to chat administrators. Returned
    only in :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    has_hidden_members: NoneBool = None
    """*Optional.* ``True``, if non-administrators can only get the list of bots
    and administrators in the chat. Returned only in
    :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    has_protected_content: NoneBool = None
    """*Optional.* ``True``, if messages from the chat can't be forwarded to
    other chats. Returned only in :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    has_visible_history: NoneBool = None
    """*Optional.* ``True``, if new chat members will have access to old
    messages; available only to chat administrators. Returned only in
    :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    sticker_set_name: NoneStr = None
    """*Optional.* For supergroups, name of group sticker set. Returned only in
    :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    can_set_sticker_set: NoneBool = None
    """*Optional.* ``True``, if the bot can change the group sticker set.
    Returned only in :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    linked_chat_id: NoneInt = None
    """*Optional.* Unique identifier for the linked chat, i.e. the discussion
    group identifier for a channel and vice versa; for supergroups and channel
    chats. Returned only in :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    location: ChatLocation | None = None
    """*Optional.* For supergroups, the location to which the supergroup is
    connected. Returned only in :meth:`~yatbaf.bot.Bot.get_chat`.
    """

    async def ban_member(
        self,
        user_id: int,
        until_date: NoneInt = None,
        revoke_messages: NoneBool = None,
    ) -> bool:
        """Ban member.

        See: :meth:`~yatbaf.bot.Bot.unban_chat_member`
        """
        return await self.bot.ban_chat_member(
            chat_id=self.id,
            user_id=user_id,
            until_date=until_date,
            revoke_messages=revoke_messages,
        )

    async def unban_member(
        self,
        user_id: int,
        only_if_banned: NoneBool = None,
    ) -> bool:
        """Unban member.

        See: :meth:`~yatbaf.bot.Bot.unban_chat_member`
        """
        return await self.bot.unban_chat_member(
            chat_id=self.id,
            user_id=user_id,
            only_if_banned=only_if_banned,
        )

    async def restrict_member(
        self,
        user_id: int,
        permissions: ChatPermissions,
        *,
        use_independent_chat_permissions: NoneBool = None,
        unitl_date: NoneInt = None,
    ) -> bool:
        """Restrict member.

        See: :meth:`~yatbaf.bot.Bot.restrict_chat_member`
        """
        return await self.bot.restrict_chat_member(
            chat_id=self.id,
            user_id=user_id,
            permissions=permissions,
            use_independent_chat_permissions=use_independent_chat_permissions,
            until_date=unitl_date,
        )

    async def promote_member(
        self,
        user_id: int,
        *,
        is_anonymous: NoneBool = None,
        can_manage_chat: NoneBool = None,
        can_post_messages: NoneBool = None,
        can_edit_messages: NoneBool = None,
        can_delete_messages: NoneBool = None,
        can_manage_video_chats: NoneBool = None,
        can_restrict_members: NoneBool = None,
        can_promote_members: NoneBool = None,
        can_change_info: NoneBool = None,
        can_invite_users: NoneBool = None,
        can_pin_messages: NoneBool = None,
        can_post_stories: NoneBool = None,
        can_edit_stories: NoneBool = None,
        can_delete_stories: NoneBool = None,
        can_manage_topics: NoneBool = None,
    ) -> bool:
        """Promote member.

        See: :meth:`~yatbaf.bot.Bot.promote_chat_member`
        """
        return await self.bot.promote_chat_member(
            chat_id=self.id,
            user_id=user_id,
            is_anonymous=is_anonymous,
            can_manage_chat=can_manage_chat,
            can_post_messages=can_post_messages,
            can_edit_messages=can_edit_messages,
            can_delete_messages=can_delete_messages,
            can_manage_video_chats=can_manage_video_chats,
            can_restrict_members=can_restrict_members,
            can_promote_members=can_promote_members,
            can_change_info=can_change_info,
            can_invite_users=can_invite_users,
            can_pin_messages=can_pin_messages,
            can_post_stories=can_post_stories,
            can_edit_stories=can_edit_stories,
            can_delete_stories=can_delete_stories,
            can_manage_topics=can_manage_topics,
        )

    async def set_andministrator_custom_title(
        self,
        user_id: int,
        custom_title: str,
    ) -> bool:
        """See: :meth:`~yatbaf.bot.Bot.set_chat_administrator_custom_title`"""
        return await self.bot.set_chat_administrator_custom_title(
            chat_id=self.id,
            user_id=user_id,
            custom_title=custom_title,
        )

    async def ban_sender_chat(self, sender_chat_id: int) -> bool:
        """See: :meth:`~yatbaf.bot.Bot.ban_chat_sender_chat`"""
        return await self.bot.ban_chat_sender_chat(
            chat_id=self.id,
            sender_chat_id=sender_chat_id,
        )

    async def unban_sender_chat(self, sender_chat_id: int) -> bool:
        """See: :meth:`~yatbaf.bot.Bot.unban_chat_sender_chat`"""
        return await self.bot.unban_chat_sender_chat(
            chat_id=self.id,
            sender_chat_id=sender_chat_id,
        )

    async def set_permissions(
        self,
        permissions: ChatPermissions,
        *,
        use_independent_chat_permissions: NoneBool = None,
    ) -> bool:
        """See: :meth:`~yatbaf.bot.Bot.set_chat_permissions`"""
        return await self.bot.set_chat_permissions(
            chat_id=self.id,
            permissions=permissions,
            use_independent_chat_permissions=use_independent_chat_permissions,
        )

    async def export_invite_link(self) -> str:
        """See: :meth:`~yatbaf.bot.Bot.export_chat_invite_link`"""
        return await self.bot.export_chat_invite_link(chat_id=self.id)

    async def create_invite_link(
        self,
        *,
        name: NoneStr = None,
        expite_date: NoneInt = None,
        member_limit: NoneInt = None,
        create_join_request: NoneBool = None,
    ) -> ChatInviteLink:
        """See: :meth:`~yatbaf.bot.Bot.create_chat_invite_link`"""
        return await self.bot.create_chat_invite_link(
            chat_id=self.id,
            name=name,
            expire_date=expite_date,
            member_limit=member_limit,
            creates_join_request=create_join_request,
        )

    async def edit_invite_link(
        self,
        invite_link: str,
        *,
        name: NoneStr = None,
        expite_date: NoneInt = None,
        member_limit: NoneInt = None,
        create_join_request: NoneBool = None,
    ) -> ChatInviteLink:
        """See: :meth:`~yatbaf.bot.Bot.edit_chat_invite_link`"""
        return await self.bot.edit_chat_invite_link(
            chat_id=self.id,
            invite_link=invite_link,
            name=name,
            expire_date=expite_date,
            member_limit=member_limit,
            creates_join_request=create_join_request,
        )

    async def revoke_invite_link(self, invite_link: str) -> ChatInviteLink:
        """See: :meth:`~yatbaf.bot.Bot.revoke_chat_invite_link`"""
        return await self.bot.revoke_chat_invite_link(
            chat_id=self.id,
            invite_link=invite_link,
        )

    async def approve_join_request(self, user_id: int) -> bool:
        """See: :meth:`~yatbaf.bot.Bot.approve_chat_join_request`"""
        return await self.bot.approve_chat_join_request(
            chat_id=self.id,
            user_id=user_id,
        )

    async def decline_join_request(self, user_id: int) -> bool:
        """See: :meth:`~yatbaf.bot.Bot.decline_chat_join_request`"""
        return await self.bot.decline_chat_join_request(
            chat_id=self.id,
            user_id=user_id,
        )

    async def set_photo(self, photo: InputFile) -> bool:
        """See: :meth:`~yatbaf.bot.Bot.set_chat_photo`"""
        return await self.bot.set_chat_photo(
            chat_id=self.id,
            photo=photo,
        )

    async def delete_photo(self) -> bool:
        """See: :meth:`~yatbaf.bot.Bot.delete_chat_photo`"""  # noqa: E501
        return await self.bot.delete_chat_photo(chat_id=self.id)

    async def set_title(self, title: str) -> bool:
        """See: :meth:`~yatbaf.bot.Bot.set_chat_title`"""
        return await self.bot.set_chat_title(
            chat_id=self.id,
            title=title,
        )

    async def set_description(self, *, description: NoneStr = None) -> bool:
        """See: :meth:`~yatbaf.bot.Bot.set_chat_description`"""
        return await self.bot.set_chat_description(
            chat_id=self.id,
            description=description,
        )

    async def pin_message(
        self,
        message_id: int,
        disable_notification: NoneBool = None,
    ) -> bool:
        """See: :meth:`~yatbaf.bot.Bot.pin_chat_message`"""
        return await self.bot.pin_chat_message(
            chat_id=self.id,
            message_id=message_id,
            disable_notification=disable_notification,
        )

    async def unpin_message(self, message_id: int) -> bool:
        """See: :meth:`~yatbaf.bot.Bot.unpin_chat_message`"""
        return await self.bot.unpin_chat_message(
            chat_id=self.id,
            message_id=message_id,
        )

    async def unpin_all_messages(self) -> bool:
        """See: :meth:`~yatbaf.bot.Bot.unpin_all_chat_messages`"""
        return await self.bot.unpin_all_chat_messages(chat_id=self.id)

    async def leave(self) -> bool:
        """See: :meth:`~yatbaf.bot.Bot.leave_chat`"""
        return await self.bot.leave_chat(chat_id=self.id)

    async def get_administrators(self) -> list[ChatMember]:
        """See: :meth:`~yatbaf.bot.Bot.get_chat_administrators`"""
        return await self.bot.get_chat_administrators(chat_id=self.id)

    async def get_member_count(self) -> int:
        """See: :meth:`~yatbaf.bot.Bot.get_chat_member_count`"""
        return await self.bot.get_chat_member_count(chat_id=self.id)

    async def get_member(self, user_id: int) -> ChatMember:
        """See: :meth:`~yatbaf.bot.Bot.get_chat_member`"""
        return await self.bot.get_chat_member(
            chat_id=self.id,
            user_id=user_id,
        )

    async def set_commands(
        self,
        commands: list[BotCommand],
        language_code: NoneStr = None,
    ) -> bool:
        """See :meth:`~yatbaf.bot.Bot.set_my_commands`,
        :class:`~yatbaf.types.bot_command_scope.BotCommandScopeChat`
        """
        return await self.bot.set_my_commands(
            commands=commands,
            scope=BotCommandScopeChat(self.id),
            language_code=language_code,
        )

    async def delete_commands(self, language_code: NoneStr = None) -> bool:
        """See :meth:`~yatbaf.bot.Bot.delete_my_commands`,
        :class:`~yatbaf.types.bot_command_scope.BotCommandScopeChat`
        """
        return await self.bot.delete_my_commands(
            scope=BotCommandScopeChat(self.id),
            language_code=language_code,
        )

    async def set_member_commands(
        self,
        user_id: int,
        commands: list[BotCommand],
        language_code: NoneStr = None,
    ) -> bool:
        """See :meth:`~yatbaf.bot.Bot.set_my_commands`,
        :class:`~yatbaf.types.bot_command_scope.BotCommandScopeChatMember`
        """
        return await self.bot.set_my_commands(
            commands=commands,
            scope=BotCommandScopeChatMember(
                chat_id=self.id,
                user_id=user_id,
            ),
            language_code=language_code,
        )

    async def delete_member_commands(
        self, user_id: int, language_code: NoneStr = None
    ) -> bool:
        """See :meth:`~yatbaf.bot.Bot.delete_my_commands`,
        :class:`~yatbaf.types.bot_command_scope.BotCommandScopeChatMember`
        """  # noqa: E501
        return await self.bot.delete_my_commands(
            scope=BotCommandScopeChatMember(
                chat_id=self.id,
                user_id=user_id,
            ),
            language_code=language_code,
        )

    async def set_administrators_commands(
        self,
        commands: list[BotCommand],
        language_code: NoneStr = None,
    ) -> bool:
        """See :meth:`~yatbaf.bot.Bot.set_my_commands`,
        :class:`~yatbaf.types.bot_command_scope.BotCommandScopeChatAdministrators`
        """  # noqa: E501
        return await self.bot.set_my_commands(
            commands=commands,
            scope=BotCommandScopeChatAdministrators(chat_id=self.id),
            language_code=language_code,
        )

    async def delete_administrators_commands(
        self, language_code: NoneStr = None
    ) -> bool:
        """See :meth:`~yatbaf.bot.Bot.delete_my_commands`,
        :class:`~yatbaf.types.bot_command_scope.BotCommandScopeChatAdministrators`
        """  # noqa: E501
        return await self.bot.delete_my_commands(
            scope=BotCommandScopeChatAdministrators(chat_id=self.id),
            language_code=language_code,
        )



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneBool

from .abc import TelegramType


@final
class ReplyKeyboardRemove(TelegramType):
    """Upon receiving a message with this object, Telegram clients will remove
    the current custom keyboard and display the default letter-keyboard. By
    default, custom keyboards are displayed until a new keyboard is sent by a
    bot. An exception is made for one-time keyboards that are hidden immediately
    after the user presses a button (see
    :class:`yatbaf.types.ReplyKeyboardMarkup`).

    See: https://core.telegram.org/bots/api#replykeyboardremove
    """

    remove_keyboard: bool
    """Requests clients to remove the custom keyboard (user will not be able to
    summon this keyboard; if you want to hide the keyboard from sight but keep
    it accessible, use ``one_time_keyboard`` in
    :class:`yatbaf.types.ReplyKeyboardMarkup`).
    """

    selective: NoneBool = None
    """*Optional.* Use this parameter if you want to remove the keyboard for
    specific users only. Targets: 1) users that are @mentioned in the text of
    the Message object; 2) if the bot's message is a reply (has
    ``reply_to_message_id``), sender of the original message.
    """



from __future__ import annotations

__all__ = ("User",)

from typing import TYPE_CHECKING
from typing import final

if TYPE_CHECKING:
    from yatbaf.types import Message


@final
class User:
    """User filter.

    Use it to filter message coming from a specific user::

        @on_message(filters=[User("user")])
        async def handler(message: Message) -> None:
            ...
    """

    __slots__ = (
        "priority",
        "users",
    )

    def __init__(self, *users: str | int, priority: int = 200) -> None:
        """
        :param users: Public username (with or without `@`) or user id.
        :param priority: Filter priority. Default 200.
        """

        self.users = frozenset([
            u.lower().removeprefix("@") if isinstance(u, str) else u
            for u in users
        ])
        self.priority = priority

    def check(self, update: Message) -> bool:
        return (u := update.from_) is not None and (
            u.id in self.users or
            (u.username is not None and u.username.lower() in self.users)
        )



import unittest.mock as mock

import pytest

from yatbaf.dispatcher import Dispatcher
from yatbaf.models import UpdateInfo
from yatbaf.router import OnMessage
from yatbaf.router import OnPoll


@pytest.mark.asyncio
async def test_resolve_no_routers(token, update_info):
    dispatcher = Dispatcher(token)
    assert not await dispatcher._resolve(update_info)


@pytest.mark.asyncio
async def test_resolve(token, handler_func, update_info):
    dispatcher = Dispatcher(
        token,
        routers=[OnMessage(handlers=[handler_func])],
    )
    await dispatcher._resolve(update_info)
    handler_func.assert_awaited_once_with(update_info.content)


@pytest.mark.asyncio
async def test_resolve_none(token, handler_func, update_info):
    dispatcher = Dispatcher(token, routers=[OnPoll(handlers=[handler_func])])
    await dispatcher._resolve(update_info)
    handler_func.assert_not_awaited()


@pytest.mark.asyncio
async def test_resolve_guard_false(token, handler_func, update_info, asyncdef):
    dispatcher = Dispatcher(
        token,
        guards=[asyncdef(False)],
        routers=[OnMessage(handlers=[handler_func])]
    )
    await dispatcher._resolve(update_info)
    handler_func.assert_not_awaited()


@pytest.mark.asyncio
async def test_resolve_guard_true(token, handler_func, update_info, asyncdef):
    dispatcher = Dispatcher(
        token,
        guards=[asyncdef(True)],
        routers=[OnMessage(handlers=[handler_func])],
    )
    await dispatcher._resolve(update_info)
    handler_func.assert_awaited_once_with(update_info.content)


@pytest.mark.asyncio
async def test_process_update(monkeypatch, token, update):
    monkeypatch.setattr(Dispatcher, "_bind_self", lambda _, v: v)  # noqa: U101
    monkeypatch.setattr(
        Dispatcher, "_resolve", resolve := mock.AsyncMock(return_value=None)
    )
    dispatcher = Dispatcher(token)
    await dispatcher.process_update(update)
    resolve.assert_awaited_once_with(
        UpdateInfo(
            id=update.update_id,
            content=update.message,
            name="message",
        )
    )



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneStr

from .abc import TelegramType


@final
class InputVenueMessageContent(TelegramType):
    """Represents the content of a venue message to be sent as the result of an
    inline query.

    See: https://core.telegram.org/bots/api#inputvenuemessagecontent
    """

    latitude: float
    """Latitude of the venue in degrees."""

    longitude: float
    """Longitude of the venue in degrees."""

    title: str
    """Name of the venue."""

    address: str
    """Address of the venue."""

    foursquare_id: NoneStr = None
    """*Optional.* Foursquare identifier of the venue, if known."""

    foursquare_type: NoneStr = None
    """*Optional.* Foursquare type of the venue, if known. (For example,
    “arts_entertainment/default”, “arts_entertainment/aquarium” or
    “food/icecream”.)
    """

    google_place_id: NoneStr = None
    """*Optional.* Google Places identifier of the venue."""

    google_place_type: NoneStr = None
    """*Optional.* Google Places type of the venue."""



from __future__ import annotations

__all__ = ("TelegramClient",)

import logging
from typing import TYPE_CHECKING
from typing import Literal
from typing import cast
from typing import final

import msgspec

from .enums import BotEnvi
from .exceptions import ChatMigratedError
from .exceptions import FileDownloadError
from .exceptions import FloodError
from .exceptions import InternalError
from .exceptions import MethodInvokeError
from .exceptions import TokenError
from .exceptions import WebhookConflictError
from .http import HttpClient
from .models import Request
from .models import ResponseError
from .types import ResponseParameters
from .utils import decode_error
from .utils import decode_response

if TYPE_CHECKING:
    from collections.abc import AsyncIterator
    from typing import Any
    from typing import Never

    from .abc import AbstractClient
    from .methods.abc import TelegramMethod
    from .models import ResponseOk
    from .types import File
    from .typing import ResultModelT

log = logging.getLogger(__name__)

SERVER_URL: Literal["https://api.telegram.org"] = "https://api.telegram.org"
"""Default api URL"""


class TelegramClient:
    """Telegram API client."""

    def __init__(
        self,
        token: str,
        *,
        api_url: str | None = None,
        environment: BotEnvi | None = None,
        client: AbstractClient | None = None,
    ) -> None:
        """
        :param token: API token.
        :param api_url: *Optional.* Api server address. Default to :attr:`SERVER_URL`.
        :param environment: *Optional.* Bot environment (see :class:`~yatbaf.enums.BotEnvi`).
        :param client: *Optional.* Http client.
        """  # noqa: E501
        self._client = HttpClient() if client is None else client
        self._api_url = SERVER_URL if api_url is None else api_url
        environment = BotEnvi.PROD if environment is None else environment
        self._creds = f"bot{token}" + (
            "/test" if environment is BotEnvi.TEST else ""
        )

    # yapf: disable
    @final
    async def invoke(
        self,
        method: TelegramMethod[ResultModelT], /, *,
        timeout: float | None = None,
        decoder: msgspec.json.Decoder[ResponseOk[ResultModelT]] | None = None,
    ) -> ResponseOk[ResultModelT]:
        """Invoke api method.

        See :ref:`methods`

        :param method: :class:`~yatbaf.methods.abc.TelegramMethod` instance.
        :param timeout: *Optional.* Request timeout.
        :param decoder: *Optional.* Response content decoder.
        """
        # yapf: enable
        url = f"{self._api_url}/{self._creds}"
        request = await self._prepare_request(url, method)
        response = await self._client.send_post(request, timeout=timeout)
        if response.status_code != 200:
            log.debug(
                f"Response status code: {response.status_code}, "
                f"content: {response.content!r}"
            )
            self._raise_for_status(response.content, method)
        return decode_response(method, response.content, decoder)

    @final
    async def download_file(
        self,
        file: File,
        chunk_size: int,
    ) -> AsyncIterator[bytes]:
        """Download file from Telegram servers.

        :param file: :class:`~yatbaf.types.file.File` instance
            (see :meth:`~yatbaf.bot.Bot.get_file`).
        :param chunk_size: Chunk length in bytes.
        """
        file_path = cast("str", file.file_path)
        url = f"{self._api_url}/file/{self._creds}/{file_path}"
        async with self._client.file_stream(url, chunk_size) as f:
            if f.status_code != 200:
                self._raise_for_status(b"".join([b async for b in f.content]))

            async for chunk in f.content:
                yield chunk

    @final
    @staticmethod
    async def _prepare_request(url: str, method: TelegramMethod) -> Request:
        """Build request.

        :param method: Method object.
        """
        content: bytes | None = None
        headers: dict[str, str] | None = None

        data, files = method._encode_params()

        # has files, data is `dict`
        if files is not None:
            data = cast("dict[str, Any]", data)
            for k, v in files.items():
                files[k] = (v.file_name, await v.read())

        # no files, data is `bytes`
        elif data is not None:
            content, data = cast("bytes", data), None
            headers = {"Content-Type": "application/json"}

        return Request(
            url=f"{url}/{method.__meth_name__}",
            content=content,
            method=method.__meth_name__,
            data=data,
            files=files,
            headers=headers,
        )

    async def close(self) -> None:
        """:meta private:"""
        log.debug("Shutting down")
        await self._client.close()

    @final
    @staticmethod
    def _raise_for_status(
        content: bytes,
        method: TelegramMethod | None = None,
    ) -> Never:
        """Raise request error. For internal use only.

        :param content: Raw response content.
        :param method: *Optional.* Method object.
        """
        response = decode_error(content)
        match response:
            case ResponseError(error_code=401):
                raise TokenError(response.description)
            case ResponseError(error_code=409):
                raise WebhookConflictError(response.description)
            # yapf: disable
            case ResponseError(parameters=ResponseParameters(retry_after=int(x))):  # noqa: E501
                raise FloodError(
                    method=cast("TelegramMethod", method),
                    error_code=response.error_code,
                    description=response.description,
                    retry_after=x,
                )
            case ResponseError(parameters=ResponseParameters(migrate_to_chat_id=int(x))):  # noqa: E501
                # yapf: enable
                raise ChatMigratedError(
                    method=cast("TelegramMethod", method),
                    error_code=response.error_code,
                    description=response.description,
                    migrate_to_chat_id=x,
                )
            case ResponseError(error_code=x) if x >= 500:
                raise InternalError(response.description)
            case _ if method is None:
                raise FileDownloadError(
                    error_code=response.error_code,
                    description=response.description,
                )

        raise MethodInvokeError(
            method=cast("TelegramMethod", method),
            error_code=response.error_code,
            description=response.description,
        )



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.types import MaskPosition


@final
class SetStickerMaskPosition(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.set_sticker_mask_position`"""

    sticker: str
    mask_position: MaskPosition | None = None



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .chat_boost import ChatBoost


@final
class UserChatBoosts(TelegramType):
    """This object represents a list of boosts added to a chat by a user.

    See: https://core.telegram.org/bots/api#userchatboosts
    """

    boosts: list[ChatBoost]
    """The list of boosts added to the chat by the user."""



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneBool

from .abc import TelegramType
from .message_entity import MessageEntity


@final
class TextQuote(TelegramType):
    """This object contains information about the quoted part of a message that
    is replied to by the given message.

    See: https://core.telegram.org/bots/api#textquote
    """

    text: str
    """Text of the quoted part of a message that is replied to by the given
    message.
    """

    position: int
    """Approximate quote position in the original message in UTF-16 code units
    as specified by the sender.
    """

    entities: list[MessageEntity] | None = None
    """*Optional.* Special entities that appear in the quote. Currently, only
    *bold*, *italic*, *underline*, *strikethrough*, *spoiler*, and
    *custom_emoji* entities are kept in quotes.
    """

    is_manual: NoneBool = None
    """Optional. ``True``, if the quote was chosen manually by the message
    sender. Otherwise, the quote was added automatically by the server.
    """



from __future__ import annotations

__all__ = ("LongPolling",)

import asyncio
import logging
import signal
from typing import TYPE_CHECKING
from typing import Any
from typing import cast

from httpx import ConnectError

from .enums import UpdateType
from .exceptions import RequestTimeoutError
from .methods import GetUpdates

if TYPE_CHECKING:
    from collections.abc import Iterable

    from .bot import Bot
    from .types import Update
    from .typing import PollingHook

log = logging.getLogger(__name__)

# TODO: windows?
SIGNALS = (signal.SIGINT, signal.SIGTERM)


class LongPolling:
    """Long polling runner."""

    __slots__ = (
        "_bot",
        "_event",
        "_method",
        "_running",
        "_on_startup",
        "_on_shutdown",
        "_connection_delay",
        "_timeout_delay",
    )

    def __init__(
        self,
        bot: Bot,
        /,
        timeout: float = 60.0,
        offset: int | None = None,
        limit: int | None = None,
        allowed_updates: list[UpdateType] | None = None,
        on_startup: Iterable[PollingHook] | None = None,
        on_shutdown: Iterable[PollingHook] | None = None,
        connection_delay: int = 20,
        timeout_delay: int = 5,
    ) -> None:
        """
        :param bot: :class:`~yatbaf.bot.Bot` isinstance.
        :param timeout: *Optional.* Timeout in seconds for long polling. Should
            be positive, short polling should be used for testing purposes only.
        :param offset: *Optional.* Identifier of the first update to be
            returned. Must be greater by one than the highest among the
            identifiers of previously received updates. By default, updates
            starting with the earliest unconfirmed update are returned. An
            update is considered confirmed as soon as getUpdates is called with
            an offset higher than its ``update_id``. The negative offset can be
            specified to retrieve updates starting from -offset update from the
            end of the updates queue. All previous updates will be forgotten.
        :param limit: *Optional.* Limits the number of updates to be retrieved.
            Values between 1-100 are accepted. Defaults to 100.
        :param allowed_updates: *Optional.* A list of the update types you want
            your bot to receive. See :class:`~yatbaf.enums.UpdateType` for a
            complete list of available update types. Specify an empty list to
            receive all update types except ``chat_member`` (default). If not
            specified, the previous setting will be used.
        :param on_startup: *Optional.* A sequence of
            :class:`~yatbaf.typing.PollingHook` to run on polling startup.
        :param on_shutdown: *Optional.* A sequence of
            :class:`~yatbaf.typing.PollingHook` to run on polling shutdown.
        :param connection_delay: *Optional.* Retry delay on request errors.
            Default 20 sec.
        :param timeout_delay: *Optional.* Retry delay on timeout error.
            Default 5 sec.


        .. note::

            ``allowed_updates`` doesn't affect updates created before the call
            to the :meth:`start`, so unwanted updates may be received for a
            short period of time.

        .. warning::

            Long polling will not work if an outgoing webhook is set up.
        """
        self._method = GetUpdates(
            timeout=timeout,
            offset=offset,
            limit=limit,
            allowed_updates=allowed_updates,
        )
        self._bot = bot
        self._event = asyncio.Event()
        self._running = False
        self._on_startup = on_startup if on_startup is not None else []
        self._on_shutdown = on_shutdown if on_shutdown is not None else []
        self._connection_delay = connection_delay
        self._timeout_delay = timeout_delay

    async def _get_updates(self, queue: asyncio.Queue) -> None:
        timeout = cast("float", self._method.timeout) + 5.0
        while not self._event.is_set():
            try:
                updates: list[Update] = (
                    await self._bot._api_client.invoke(
                        self._method,
                        timeout=timeout,
                    )
                ).result

                if updates:
                    self._method.offset = updates[-1].update_id + 1
                    queue.put_nowait(updates)

            except (ConnectError, RequestTimeoutError) as error:
                if isinstance(error, RequestTimeoutError):
                    message = (
                        "Connection timed out. "
                        f"Next try in {self._timeout_delay} sec."
                    )
                    delay = self._timeout_delay
                else:
                    message = (
                        "Network connection error. "
                        f"Next try in {self._connection_delay} sec."
                    )
                    delay = self._connection_delay

                log.error(message)
                await asyncio.sleep(delay)

            except Exception as error:
                self._event.set()
                await self._bot._error_handler.on_error(error)

    async def _main_loop(self) -> None:
        log.info("Starting long polling...")
        queue: asyncio.Queue = asyncio.Queue()
        updates_task = asyncio.create_task(self._get_updates(queue))

        bot = self._bot
        while not self._event.is_set():
            try:
                for update in (await asyncio.wait_for(queue.get(), 0.2)):
                    bot.tasks._create_task(
                        bot.process_update(update),
                        name=f"update-task-{update.update_id}",
                    )
            except TimeoutError:
                pass

        await bot.tasks.kill(updates_task)

    async def _startup(self) -> None:
        log.debug("Sturtup...")
        if self._running:
            return
        self._event.clear()

        for func in self._on_startup:
            await func(self._bot)

        loop = asyncio.get_running_loop()
        for s in SIGNALS:
            loop.add_signal_handler(s, self.stop)
        self._running = True

    async def _shutdown(self) -> None:
        log.debug("Shutdown...")
        if not self._running:
            return
        self._running = False

        await self._bot.tasks.wait_all()
        try:
            for func in self._on_shutdown:
                await func(self._bot)
        finally:
            await self._bot.shutdown()

    async def _run(self) -> None:
        try:
            await self._startup()
        except Exception as error:
            log.error("Startup failed.", exc_info=error)
            return

        await self._main_loop()
        await self._shutdown()

    def start(self) -> None:
        """Start long polling"""
        asyncio.run(self._run())

    def stop(self, *args: Any) -> None:  # noqa: U100
        """Stop long polling"""
        self._event.set()



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import Message

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt


@final
class ForwardMessage(TelegramMethod[Message]):
    """See :meth:`yatbaf.bot.Bot.forward_message`"""

    chat_id: str | int
    from_chat_id: int | str
    message_id: int
    message_thread_id: NoneInt = None
    disable_notification: NoneBool = None
    protect_content: NoneBool = None



import re
import unittest.mock as mock

import pytest

from yatbaf import bot
from yatbaf import methods


def capitalize(w):
    return f"{w[0].upper()}{w[1:]}"


def snake_case(s):
    return '_'.join(
        re.sub(
            '([A-Z][a-z]+)',
            r' \1',
            re.sub('([A-Z]+)', r' \1', s.replace('-', ' '))
        ).split()
    ).lower()


@pytest.fixture
def api_methods(api_spec):
    return api_spec["methods"]


def test_method_object(api_methods):
    for api_method in api_methods.values():
        assert hasattr(methods, capitalize(api_method["name"]))


def test_method_object_fields(api_methods):
    for api_method in api_methods.values():
        method_obj = getattr(methods, capitalize(api_method["name"]))
        for field in api_method.get("fields", []):
            assert hasattr(method_obj, field["name"])


@pytest.mark.asyncio
async def test_method_func(monkeypatch, api_methods, token):
    monkeypatch.setattr(bot.Bot, "_call_api", mock.AsyncMock())
    bot_obj = bot.Bot(token)
    for api_method in api_methods.values():
        with mock.patch(f"yatbaf.bot.{capitalize(api_method['name'])}") as o:
            method = getattr(bot_obj, snake_case(api_method["name"]))
            fields = {}
            for i, field in enumerate(api_method.get("fields", [])):
                fields[field["name"]] = i
            await method(**fields)
            if fields:
                o.assert_called_once_with(**fields)
            else:
                o.assert_called_once()



import pytest

from yatbaf.dispatcher import Dispatcher
from yatbaf.exceptions import BotWarning
from yatbaf.router import OnMessage
from yatbaf.router import OnPoll


def test_add_router():
    router = OnMessage()
    nested = OnMessage()
    router.add_router(nested)
    assert len(router._routers) == 1
    assert router._routers[0] is nested
    assert nested._parent is router


def test_add_router_self(router):
    with pytest.raises(ValueError):
        router.add_router(router)


def test_add_router_wrong_type(router):
    with pytest.raises(ValueError):
        router.add_router(OnPoll())


def test_add_router_registered(router):
    nested = OnMessage()
    router.add_router(nested)
    router2 = OnMessage()
    with pytest.raises(ValueError):
        router2.add_router(nested)


def test_add_router_registered_same_router(router):
    nested = OnMessage()
    router.add_router(nested)
    with pytest.warns(BotWarning):
        router.add_router(nested)


def test_add_router_dispatcher(router, token):
    dispatcher = Dispatcher(token)
    with pytest.raises(RuntimeError):
        router.add_router(dispatcher)


def test_add_router_duplicate_same_obj():
    router = OnMessage()
    nested = OnMessage()
    router.add_router(nested)
    with pytest.warns(BotWarning):
        router.add_router(nested)
    assert len(router._routers) == 1
    assert router._routers[0] is nested


def test_add_router_duplicate_equal_obj():
    router = OnMessage()
    nested1 = OnMessage(name="router")
    nested2 = OnMessage(name="router")
    router.add_router(nested1)
    with pytest.warns(BotWarning):
        router.add_router(nested2)
    assert len(router._routers) == 1
    assert router._routers[0] is nested1



import pytest

from yatbaf.methods import AddStickerToSet
from yatbaf.methods import Close
from yatbaf.methods import CreateNewStickerSet
from yatbaf.methods import EditMessageMedia
from yatbaf.methods import GetMe
from yatbaf.methods import GetUpdates
from yatbaf.methods import SendAnimation
from yatbaf.methods import SendAudio
from yatbaf.methods import SendDocument
from yatbaf.methods import SendMediaGroup
from yatbaf.methods import SendPhoto
from yatbaf.methods import SendSticker
from yatbaf.methods import SendVideo
from yatbaf.methods import SendVideoNote
from yatbaf.methods import SendVoice
from yatbaf.methods import SetStickerSetThumbnail
from yatbaf.methods import SetWebhook
from yatbaf.methods import StopMessageLiveLocation
from yatbaf.methods import UploadStickerFile
from yatbaf.methods.abc import TelegramMethodWithFile
from yatbaf.methods.abc import TelegramMethodWithMedia
from yatbaf.types import Message
from yatbaf.types import Update
from yatbaf.types import User


def test_base_subclass_attrs():
    assert not hasattr(TelegramMethodWithMedia, "__meth_name__")
    assert not hasattr(TelegramMethodWithMedia, "__meth_result_model__")

    assert not hasattr(TelegramMethodWithFile, "__meth_name__")
    assert not hasattr(TelegramMethodWithFile, "__meth_result_model__")


def test_attrs():
    assert hasattr(GetMe, "__meth_name__")
    assert hasattr(GetMe, "__meth_result_model__")


def test_file_fields():
    assert SendAnimation.__meth_file_fields__ == ("animation", "thumbnail")
    assert SendAudio.__meth_file_fields__ == ("audio", "thumbnail")
    assert SendDocument.__meth_file_fields__ == ("document", "thumbnail")
    assert SendPhoto.__meth_file_fields__ == ("photo",)
    assert SendSticker.__meth_file_fields__ == ("sticker",)
    assert SendVideo.__meth_file_fields__ == ("video", "thumbnail")
    assert SendVideoNote.__meth_file_fields__ == ("video_note", "thumbnail")
    assert SendVoice.__meth_file_fields__ == ("voice",)
    assert UploadStickerFile.__meth_file_fields__ == ("sticker",)
    assert SetWebhook.__meth_file_fields__ == ("certificate",)
    assert SetStickerSetThumbnail.__meth_file_fields__ == ("thumbnail",)


def test_media_fields():
    assert AddStickerToSet.__meth_media_fields__ == ("sticker",)
    assert SendMediaGroup.__meth_media_fields__ == ("media",)
    assert EditMessageMedia.__meth_media_fields__ == ("media",)
    assert CreateNewStickerSet.__meth_media_fields__ == ("stickers",)


def test_method_name():
    assert GetMe.__meth_name__ == "getme"
    assert str(GetMe()) == "getme"
    assert GetMe()._get_name() == "getme"


@pytest.mark.parametrize(
    "method,model",
    (
        (GetMe, User),
        (Close, bool),
        (GetUpdates, list[Update]),
        (StopMessageLiveLocation, Message | bool),
    ),
)
def test_method_result_model(method, model):
    assert method.__meth_result_model__ == model
    assert method._get_result_model() == model



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from .abc import TelegramType


@final
class WebhookInfo(TelegramType):
    """Describes the current status of a webhook.

    See: https://core.telegram.org/bots/api#webhookinfo
    """

    url: str
    """Webhook URL, may be empty if webhook is not set up."""

    has_custom_certificate: bool
    """``True``, if a custom certificate was provided for webhook certificate
    checks.
    """

    pending_update_count: int
    """Number of updates awaiting delivery."""

    ip_address: NoneStr = None
    """*Optional.* Currently used webhook IP address."""

    last_error_date: NoneInt = None
    """*Optional.* Unix time for the most recent error that happened when trying
    to deliver an update via webhook.
    """

    last_error_message: NoneStr = None
    """*Optional.* Error message in human-readable format for the most recent
    error that happened when trying to deliver an update via webhook.
    """

    last_synchronization_error_date: NoneInt = None
    """*Optional.* Unix time of the most recent error that happened when trying
    to synchronize available updates with Telegram datacenters.
    """

    max_connections: NoneInt = None
    """*Optional.* The maximum allowed number of simultaneous HTTPS connections
    to the webhook for update delivery.
    """

    allowed_updates: list[str] | None = None
    """*Optional.* A list of update types the bot is subscribed to. Defaults
    to all update types except chat_member.
    """
